{-# LANGUAGE CPP, BangPatterns, MagicHash, UnboxedTuples, DeriveDataTypeable #-}
# if __GLASGOW_HASKELL__ >= 612
{-# OPTIONS_GHC -no-auto-all -no-auto -no-caf-all #-}
# endif
module Game(Game, putLogCommand, primRunGame, equivalently, finish, getGS, setGS, modifyGS) where

import Char

import Control.Exception
import Data.Typeable
import Data.IORef
import GHC.Prim
import GHC.IO
# if __GLASGOW_HASKELL__ < 612
import GHC.IOBase
# endif

newtype Game a = Game { unGame :: GameState -> IORef [LogCommand] -> State# RealWorld -> (# State# RealWorld, GameState, a #) }

data GameFinishedException = GFE GameState GameResult
  deriving (Typeable, Show, Eq)

data GameInconsistentException = GIE GameState GameState GameState
  deriving (Typeable, Show)

instance Exception GameFinishedException
instance Exception GameInconsistentException

instance Monad Game where
  return x = Game $ \gs _ k -> (# k, gs, x #)
  Game f >>= g = Game $ \gs ref rs -> case f gs ref rs of
    (# rs1, gs1, v #) -> unGame (g v) gs1 ref rs1

putLogCommand :: LogCommand -> Game ()
putLogCommand lc = Game $ \gs ref rs -> case unIO (readIORef ref) rs of
  (# rs1, log #) -> case unIO (writeIORef ref (lc:log)) rs1 of
    (# rs2, _ #) -> (# rs2, gs, () #)

primRunGame :: Game a -> GameState -> (Either GameResult a, [LogCommand], GameState)
primRunGame (Game x) gs = unsafePerformIO run
  where
    run = do
      v <- newIORef []
      r <- try $ body v
      log <- readIORef v
      let 
        (e, gs') = case r of
          Left (GFE gs' gr) -> (Left gr, gs')
          Right (x, gs') -> (Right x, gs')
      return (e, reverse log, gs')

    body v = IO $ \rs -> case x gs v rs of
      (# rs1, gs1, x #) -> (# rs1, (x, gs1) #)

equivalently :: (Eq a) => Game a -> Game a -> Game a
equivalently x y = Game $ \gs ref rs -> case g_try x gs ref rs of
  (# rs1, gs1, r1 #) -> case unIO (newIORef []) rs1 of
    (# rs2, sink #) -> case g_try y gs sink rs2 of
      (# rs3, gs2, r2 #)
        | gs1 == gs2 && r1 == r2 -> untry r1 gs1 rs3
        | otherwise -> unGame (ioToGame $ throwIO $ GIE gs gs1 gs2) gs2 ref rs3
  where
    untry :: Either GameFinishedException a -> GameState -> State# RealWorld -> (# State# RealWorld, GameState, a #)
    untry val gs rs = case val of
      Left gfi -> unGame (ioToGame $ throwIO gfi) undefined undefined rs
      Right x -> (# rs, gs, x #)

    g_try :: Game a -> GameState -> IORef [LogCommand] -> State# RealWorld -> (# State# RealWorld, GameState, Either GameFinishedException a #)
    g_try (Game f) gs ref rs = case unIO (try io) rs of
      (# rs1, v #) -> case v of
        Left e -> (# rs1, gs, Left e #)
        Right (gs1, x) -> (# rs1, gs1, Right x #)
      where
        io = IO $ \rs -> case f gs ref rs of
          (# rs1, gs1, v #) -> (# rs1, (gs1, v) #)

finish :: GameResult -> Game a
finish gr = do
  gs <- getGS
  ioToGame $ throwIO $ GFE gs gr

ioToGame :: IO a -> Game a
ioToGame (IO x) = Game $ \gs _ rs -> case x rs of
  (# rs1, v #) -> (# rs1, gs, v #)

getGS :: Game GameState
getGS = Game $ \gs _ rs -> (# rs, gs, gs #)

setGS :: GameState -> Game ()
setGS gs = Game $ \_ _ rs -> case gs of !rgs -> (# rs, rgs, () #)

modifyGS :: (GameState -> GameState) -> Game ()
modifyGS f = Game $ \gs _ rs -> case f gs of !rgs -> (# rs, rgs, () #)
