{-# LANGUAGE BangPatterns #-}
module GameUtil where

import Char
import qualified Bag as B
import Game
import qualified CharRef as CR

import Control.Monad
import Data.List
import qualified Data.Map as M
import Data.Maybe

addSideEffect
  :: Int
  -> (GameState -> CharRef -> Game ())
  -> ActionChain
  -> ActionChain
addSideEffect !phase sideEffect action = add phase [] action
  where
    add 0 extraDamaged (AC a) = AC $ \gs -> do
      r@(damaged, _) <- a gs
      mapM_ (sideEffect gs) $ extraDamaged ++ damaged
      return r
    add d extraDamaged (AC a) = AC $ \gs -> do
      (damaged, next) <- a gs
      return (damaged, Just $
        add (d-1) (extraDamaged++damaged) (fromMaybe noAC next))

    noAC = AC $ const $ return ([], Nothing)

damageDeclarationPhase, damageDealingPhase, destructionPhase, pressurizationPhase, soshitePhase :: Int
damageDeclarationPhase = 0
damageDealingPhase = 1
destructionPhase = 2
pressurizationPhase = 3
soshitePhase = 4

putStun, putPoison :: GameState -> CharRef -> Game ()
putStun = putBad "スタン" sbStun $ \b -> b{ sbStun = sbStun b + 1 }
putPoison = putBad "毒" sbPoison $ \b -> b{ sbPoison = sbPoison b + 1 }

putBad :: String -> (StatusBad -> Int) -> (StatusBad -> StatusBad) -> GameState -> CharRef -> Game ()
putBad name getter adder gs y = do
  cs <- getCS y
  let sb = csStatusBad cs
  let sb' = adder sb
  mes $ show y ++ "は" ++ name ++ "を得た。" ++ show (getter sb) ++ " -> " ++ show (getter sb')
  setCS y $ cs{ csStatusBad = sb' }

destroySword :: GameState -> CharRef -> Game ()
destroySword gs y = case csSwords $ gsChars gs CR.! y of
  [] -> return ()
  target:_ -> do
    cs <- getCS y
    mes $ show y ++ "の腰骨が折れた"
    mes $ show y ++ "の" ++ (show $ ssSword $ csSwordState cs M.! target) ++ "は破壊された"
    setCS y $ deleteSword target cs

setHP :: CharRef -> Integer -> Game ()
setHP cr n = do
  cs <- getCS cr
  setCS cr $ cs{ csHp = n }
  when (n <= 0 && csHp cs > 0) $ kill cr
  when (csHp cs <= 0 && n > 0) $ revive cr

damage :: String -> Integer -> CharRef -> Game ()
damage desc value cr
  | value <= 0 = return ()
  | otherwise = do
    cs <- getCS cr
    let newhp = csHp cs - value
    mes $ show cr ++ "は" ++ desc ++ "によって" ++ show value ++ "点のダメージを受けた。" ++ hpChange (csHp cs) newhp
    setHP cr newhp

revive :: CharRef -> Game ()
revive cr = do
  cs <- getCS cr
  when (csDead cs) $ do
    mes $ show cr ++ "は復活した"
    setCS cr cs{ csDead = False }

kill :: CharRef -> Game ()
kill cr = do
  n <- getTurnCount
  when (n /= 0) $ reallyKill cr

reallyKill :: CharRef -> Game ()
reallyKill cr = do
  mes $ show cr ++ "は死んだ"
  revived <- invokeRevive cr
  when (not revived) $ do
    modifyCS cr $ \cs -> cs{ csDead = True }
    modifyGS $ \gs -> gs{ gsSomeoneMayBeDead = True }

invokeRevive :: CharRef -> Game Bool
invokeRevive cr = do
  cs <- getCS cr
  case B.pop $ csStatusRevive cs of
    Just (Soul, sr) -> activate cs sr "ソウル" (csMaxHp cs)
    Just (Life, sr) -> activate cs sr "ライフ" 1
    Nothing -> return False
  where
    activate cs sr reason val = do
      mes $ reason ++ "が発動した。" ++ hpChange (csHp cs) val
      setCS cr cs{ csHp = val, csStatusRevive = sr }
      if val <= 0
        then invokeRevive cr
        else return True

hpChange :: Integer -> Integer -> String
hpChange from to = "HP: " ++ show from ++ " -> " ++ show to

getTurnCount :: Game Int
getTurnCount = liftM gsTurnCount getGS

initialSS :: Sword -> SwordState
initialSS s = SS
  { ssSword = s
  , ssUsed = False
  , ssStun = 0
  , ssPoison = 0
  , ssDestruction = 0
  , ssAll = False
  , ssEarly = 0
  , ssLimitAdded = False
  , ssForcedReactsToDanger = False
  }

noBadStatus :: StatusBad
noBadStatus = SB
  { sbPoison = 0
  , sbBadPoison = 0
  , sbStun = 0
  , sbAbortion = 0
  , sbSleep = 0
  }

getECNext :: GameState -> EffectContext -> Maybe (SwordRef, SwordState)
getECNext gs ec = do
  sw <- ecNextSword ec
  ss <- M.lookup sw $ csSwordState cs
  return (sw, ss)
  where
    cs = gsChars gs CR.! ecOwner ec

teamcase :: a -> a -> TeamRef -> a
teamcase x y t = case t of
  Team0 -> x
  Team1 -> y

opponent :: TeamRef -> TeamRef
opponent = teamcase Team1 Team0

gsTeam :: TeamRef -> GameState -> [CharRef]
gsTeam tr = teamcase gsTeam0 gsTeam1 tr

incrementTurnCount = do
  n <- getTurnCount
  putLogCommand $ PutNoIndent $ "<<< Turn " ++ show (n + 1) ++ " >>>"
  when (n == 0) bringDeath -- The blessed 0th turn is over; now all swordsmen are mortal.
  setTurnCount (n + 1)
  handlePeriod

bringDeath = do
  ccs <- presentCharsWithState
  let dead = map fst $ filter ((<=0) . csHp . snd) ccs
  mapM_ reallyKill dead

presentCharsWithState :: Game [(CharRef, CharacterState)]
presentCharsWithState = do
  gs <- getGS
  return $ map (\c -> (c, gsChars gs CR.! c)) $ gsTeam0 gs ++ gsTeam1 gs

handlePeriod = do
  n <- getTurnCount
  forEachCharacter $ \ch -> do
    cs <- getCS ch
    let to_remove = filter (maybe False (<n) . swPeriod . ssSword . snd) $ M.toList $ csSwordState cs
    when (not $ null to_remove) $ 
      putLogCommand $ PutNoIndent $ show ch ++ "の" ++ (intercalate "、" $ map (show . snd) to_remove) ++ "はピリオドによって消滅した"
    setCS ch $ foldl' (flip deleteSword) cs $ map fst to_remove

setTurnCount :: Int -> Game ()
setTurnCount n = modifyGS $ \gs -> gs{ gsTurnCount = n }

getCS :: CharRef -> Game CharacterState
getCS ref = do
  gs <- getGS
  case CR.lookup ref $ gsChars gs of
    Just c -> return c

modifyCS :: CharRef -> (CharacterState -> CharacterState) -> Game ()
modifyCS ref f = do
  gs <- getGS
  setGS $ gs{ gsChars = CR.update (Just . f) ref $ gsChars gs }

setCS :: CharRef -> CharacterState -> Game ()
setCS ref cs = do
  gs <- getGS
  setGS $ gs{ gsChars = CR.insert ref cs $ gsChars gs }

mes :: String -> Game ()
mes str = seq (length str) $ putLogCommand $ PutLine str

indented :: Game a -> Game a
indented f = do
  putLogCommand $ ChangeIndent 2
  v <- f
  putLogCommand $ ChangeIndent (-2)
  return v

crTeam :: CharRef -> TeamRef
crTeam (CR n _) = if even n then Team0 else Team1

deleteSword :: SwordRef -> CharacterState -> CharacterState
deleteSword sr cs = cs{ csSwords = filter (/=sr) $ csSwords cs, csSwordState = M.delete sr $ csSwordState cs }

addSword :: CharRef -> SwordState -> Game SwordRef
addSword cr ss = do
  cs <- getCS cr
  let supply = csSwordRefSupply cs
  let sr = SR supply
  setCS cr $ cs
    { csSwordState = M.insert sr ss $ csSwordState cs
    , csSwordRefSupply = supply + 1
    , csSwords = csSwords cs ++ [sr]
    }
  return sr

forEachCharacter :: (CharRef -> Game ()) -> Game ()
forEachCharacter f = do
  gs <- getGS
  mapM_ f $ gsTeam0 gs ++ gsTeam1 gs

csSide :: CharacterState -> TeamRef
csSide = ciSide . csInfo

primary0, primary1 :: CharRef
primary0 = CR 0 Nothing
primary1 = CR 1 Nothing

registerSwordTransfer :: CharRef -> SwordRef -> CharRef -> SwordRef -> Game ()
registerSwordTransfer oldc olds newc news = modifyGS $ \gs ->
  gs{ gsSwordTransfer = M.insert (oldc, olds) (newc, news) $ gsSwordTransfer gs }

clearSwordTransfer :: Game ()
clearSwordTransfer = modifyGS $ \gs -> gs{ gsSwordTransfer = M.empty }
