module Match(match, testMatch) where

import qualified Bag as B
import Char
import Util
import GameUtil
import Game
import qualified CharRef as CR

import Data.Monoid
import Data.Either
import Data.Maybe
import Data.Function
import Data.List
import Control.Monad
import qualified Data.Map as M

initialState :: Character -> Character -> GameState
initialState c0 c1 = GS
  { gsChars = CR.fromList [(cr0, initialCS $ mk_info d0 Team0 c0), (cr1, initialCS $ mk_info d1 Team1 c1)]
  , gsTeam0 = [cr0]
  , gsTeam1 = [cr1]
  , gsUniqueSupply0 = 1
  , gsUniqueSupply1 = 1
  , gsTurnCount = 0
  , gsSomeoneMayBeDead = False
  , gsSwordTransfer = M.empty
  }
  where
    cr0 = CR 0 (Just d0) -- these magic numbers should match the definition of primary[01]
    cr1 = CR 1 (Just d1)
    n0 = filter (/='/') $ chName c0
    n1 = filter (/='/') $ chName c1
    (d0, d1)
      | n0 == n1 = (n0 ++ "/0", n1 ++ "/1")
      | otherwise = (n0, n1)
    mk_info disp team orig = CI{ ciDispName = disp, ciSide = team, ciBuildRestriction = Nothing, ciOriginal = orig }

initialCS :: CharacterInfo -> CharacterState
initialCS ci = CS
  { csInfo = ci
  , csDead = False
  , csMaxHp = chHp c
  , csHp = chHp c
  , csAtk = chAtk c
  , csDef = chDef c
  , csSpd = chSpd c
  , csSwords = zipWith const (map SR [0..]) $ chSwords c
  , csAp = 0
  , csReservedAp = 0
  , csMp = 0
  , csStatusBad = noBadStatus
  , csStatusRevive = B.empty
  , csStatusDefence = B.empty
  , csStatusGuard = B.empty
  , csStatusOther = noOtherStatus
  , csSwordState = M.fromList $ zip (map SR [0..]) $ map initialSS $ chSwords c 
  , csSwordRefSupply = length $ chSwords c
  }
  where
    c = ciOriginal ci

noOtherStatus :: StatusOther
noOtherStatus = SO
  { soSlow = 0
  , soDrive = 0
  , soQuick = 0
  , soPush = 0
  , soHeavensDoor = False
  , soProvoke = False
  }

match :: Character -> Character -> (GameResult, [String], GameState)
match c0 c1 = runGame c0 c1 gameBody

testMatch :: Character -> Character -> IO ()
testMatch c0 c1 = do
  print gr
  putStr $ unlines msgs
  print gs
  where
    (gr, msgs, gs) = match c0 c1

runGame :: Character -> Character -> Game a -> (GameResult, [String], GameState)
runGame c0 c1 g = (gr, renderLogCommands out, gs)
  where
    (e, out, gs) = primRunGame g (initialState c0 c1)
    gr = case e of
      Left gr -> gr
      Right _ -> error "runGame: game not finished"

renderLogCommands :: [LogCommand] -> [String]
renderLogCommands = process 0
  where
    process n _
      | seq n False = undefined
    process n [] = []
    process n (x:xs) = case x of
      PutLine ln -> shift n ln : process n xs
      ChangeIndent k -> process (n+k) xs
      PutNoIndent ln -> ln : process n xs

    shift 0 xs = xs
    shift n xs = ' ' : shift (n-1) xs

gameBody :: Game a
gameBody = forever turn

turn :: Game ()
turn = do
  checkpoint
  showCurrentState
  statusEffectPhase
  resetAp
  doAction
  cleanup
  incrementTurnCount

showCurrentState = do
  gs <- getGS
  let cs = gsChars gs
  let multi = length (gsTeam0 gs ++ gsTeam1 gs) > 2
  let br = when multi $ mes ""
  br
  mapM_ (mes . show . (cs CR.!)) $ gsTeam0 gs
  br
  mapM_ (mes . show . (cs CR.!)) $ gsTeam1 gs
  mes ""

statusEffectPhase = applyBadPoison >> applyPoison

applyBadPoison = transformAllPresent $ \cid cs -> let
  sb = csStatusBad cs
  bp = sbBadPoison sb
  sb' = sb{ sbPoison = sbPoison sb + bp }
  in (cs{ csStatusBad = sb' }, replicate bp $ show cid ++ "は猛毒によって毒を得た")

applyPoison :: Game ()
applyPoison = do
  gs <- getGS
  forM_ (gsTeam0 gs ++ gsTeam1 gs) $ \cid -> do
    cs <- getCS cid
    damage "毒" (toInteger $ sbPoison $ csStatusBad cs) cid
  checkpoint

resetAp :: Game ()
resetAp = transformAllPresent $ \cid cs -> let
  sb = csStatusBad cs
  sleep = sbSleep sb
  in if sleep > 0
    then (cs{ csAp = 0, csReservedAp = 0, csStatusBad = sb{ sbSleep = sleep - 1 } }, [show cid ++ "は眠っている"])
    else (cs{ csAp = csSpd cs }, [])

doAction :: Game ()
doAction = repeatedly action1
  where
    action1 = do
      rs <- actives
      case maxesOn priority rs of
        [] -> return False
        xs -> do
          simulAction xs'
          clearSwordTransfer
          checkpoint
          return True
          where
            xs' = [x0, x1]
            (x0, x1) = span ((==Team0) . crTeam . fst) xs

    priority (_, cs) = (soQuick so - soSlow so, csSpd cs)
      where
        so = csStatusOther cs

    actives = do
      t <- getTurnCount
      pcs <- presentCharsWithState
      return $ filter (isActive t . snd) pcs

-- does the character have a right to do something?
isActive :: Int -> CharacterState -> Bool
isActive t cs = csAp cs > 0 || csReservedAp cs > 0 && canUseSwordOrStun t cs

-- can the character do something other than nop?
canUseSwordOrStun :: Int -> CharacterState -> Bool
canUseSwordOrStun t cs = fst (chooseAction t cs) /= ATNop

-- the argument list must be ordered so that the order of characters within each team is preserved.
simulAction :: [[(CharRef, CharacterState)]] -> Game ()
simulAction pss = do
  t <- getTurnCount
  pssa <- forM pss $ \ps -> forM ps $ \(cr, cs) -> case chooseAction t cs of
    (act, cs') -> do
      setCS cr $ if csAp cs' > 0
        then cs'{ csAp = csAp cs' - 1 }
        else cs'{ csReservedAp = csReservedAp cs' - 1 }
      return (cr, act)
  let psa = concat pssa
  if all ((==ATNop) . snd) psa
    then everyoneDoesNothing $ map fst psa
    else do
      mes $ makeActionStatement psa
      let [t0, t1] = map (revSortGroupOn (actionPriority . snd)) pssa
      ats <- attackTargets
      indented $ mapM_ (process ats) $ concatMap (uncurry zipMaybe) $ merge t0 t1
  where
    order xs = sortBy (flip compare `on` actionPriority . snd . snd) $ zip [0..] xs

    process ats (Just x, Just y) = superSimulAction ats x y
    process ats (Just x, Nothing) = singleAction ats multi Team0 x
    process ats (Nothing, Just y) = singleAction ats multi Team1 y

    merge xs [] = zip xs $ repeat []
    merge [] ys = zip (repeat []) ys
    merge xxs@(x:xs) yys@(y:ys) = case (compare `on` actionPriority . snd . head) x y of
      GT -> (x, []) : merge xs yys
      LT -> ([], y) : merge xxs ys
      EQ -> (x, y) : merge xs ys

    multi = not $ null $ drop 1 $ concat pss

makeActionStatement :: [(CharRef, ActionType)] -> String
makeActionStatement psa = concat $ zipWith state psa $ map (const False) (tail psa) ++ [True]
  where
    state (cp, ac) last = actionState last (show cp) ac

everyoneDoesNothing :: [CharRef] -> Game ()
everyoneDoesNothing crs = do
  gs <- getGS
  let cr_cs = map (pair (gsChars gs CR.!)) crs
  let ex = minimum $ map (csAp . snd) cr_cs -- no need to consider reserved ap, because they are never used to do NOP
  let msgtail = if ex == 0 then "" else "。" ++ show (ex+1) ++ "回。"
  mes $ makeActionStatement (map (flip (,) ATNop) crs) ++ msgtail
  when (ex > 0) $ forM_ cr_cs $ \(cr, cs) -> setCS cr cs{ csAp = csAp cs - ex }

data ActionType 
  = ATStun
  | ATUse SwordRef (Maybe SwordRef) SwordState
  | ATNop
  deriving Eq

actionPriority :: ActionType -> Int
actionPriority at = case at of
  ATStun -> 7
  ATUse _ _ ss -> case swEffect $ ssSword ss of
    EAux{} -> 6
    ECtl{} -> 5
    EAdd{} -> 4
    EAtk{} -> 3
    EBld{} -> 2
    EEnv{} -> 1
  ATNop -> 0

actionDescr :: String -> ActionType -> String
actionDescr name at = name ++ "の" ++ case at of
  ATStun -> "スタン処理"
  ATUse _ _ ss -> (swName $ ssSword ss)
  ATNop -> "無行動"

actionState :: Bool -> String -> ActionType -> String
actionState terminal name at = name ++ "は" ++ case at of
  ATStun -> "気絶して" ++ p "おり" "いる"
  ATUse _ _ ss -> show ss ++ "を使" ++ p "い" "った"
  ATNop -> "何も" ++ p "せず" "しない"
  where
    p x y = if terminal then y else x ++ "、"

chooseAction :: Int -> CharacterState -> (ActionType, CharacterState)
chooseAction turn cs
  | sbStun sb > 0 = (ATStun, cs{ csStatusBad = sb{ sbStun = sbStun sb - 1 } })
  | otherwise = case filter (usable . head) $ filter (not . null) $ tails $ csSwords cs of
    k:_ -> (ATUse x next ss, cs'')
      where
        cs''
          | ssLimited ss = deleteSword x cs'
          | otherwise = cs'{ csSwordState = M.insert x ss{ ssUsed = True } $ csSwordState cs' }
        cs' = case swMP $ ssSword ss of
          Just k -> cs{ csMp = csMp cs - k }
          Nothing -> cs
        ss = sw x
        x = head k
        next = case tail k of
          [] -> Nothing
          a:_ -> Just a
    [] -> (ATNop, cs)
  where
    sb = csStatusBad cs
    sw x = csSwordState cs M.! x
    usable t
      =  not (ssUsed ss)
      && turn >= ssAvailableTurn ss
      && maybe True (csMp cs>=) (swMP $ ssSword ss)
      && (not (ssReactsToDanger ss) || csHp cs <= (div (csMaxHp cs) 2))
      && (not (ssReactsToIllness ss) || csStatusBad cs /= noBadStatus)
      where
        ss = sw t

attackTargets :: Game (CharRef, CharRef)
attackTargets = do
  gs <- getGS
  return (f gsTeam1 gs, f gsTeam0 gs)
  where
    f members gs = head $ filter (soProvoke . csStatusOther . (m CR.!)) mem ++ mem
      where
        mem = members gs
        m = gsChars gs

superSimulAction :: (CharRef, CharRef) -> (CharRef, ActionType) -> (CharRef, ActionType) -> Game ()
superSimulAction ats (cr0, at0) (cr1, at1) = case (at0, at1) of
  (ATStun, ATStun) -> return ()
  (ATNop, ATNop) -> return ()
  (ATUse sr0 next0 ss0, ATUse sr1 next1 ss1) -> do
    mes $ actionDescr (show cr0) at0 ++ "と" ++ actionDescr (show cr1) at1 ++ "との超同時処理"
    indented $ case (ssEffect ss0, ssEffect ss1) of
      (EAtk re0, EAtk re1) -> performAttack ats (Just (ec0, re0)) (Just (ec1, re1))
      (EAux re0, EAux re1) -> performAux (Just (ec0, re0)) (Just (ec1, re1))
      (ECtl a0 x0, ECtl a1 x1) -> do
        gs <- getGS
        runCtl gs ec0 a0 x0
        runCtl gs ec1 a1 x1
      (EAdd x0, EAdd x1) -> do
        gs <- getGS
        runAdd gs ec0 x0
        runAdd gs ec1 x1
      (EBld r0 x0, EBld r1 x1) -> do
        gs <- getGS
        build r0 cr0 Team0 (x0 gs ec0)
        build r1 cr1 Team1 (x1 gs ec1)
      (EEnv x0, EEnv x1) -> do
        gs <- getGS
        x0 gs ec0
        x1 gs ec1
    where
      ec0 = makeEC cr0 sr0 next0
      ec1 = makeEC cr1 sr1 next1

singleAction :: (CharRef, CharRef) -> Bool -> TeamRef -> (CharRef, ActionType) -> Game ()
singleAction ats needMsg team (cr, at) = case at of
  ATStun -> return ()
  ATNop -> return ()
  ATUse sr next ss
    | needMsg -> do
      when needMsg $ mes $ "処理: " ++ actionDescr (show cr) at
      indented body
    | otherwise -> body
    where
      body = case ssEffect ss of
        EAtk re -> withRE (performAttack ats) re ec
        EAux re -> withRE performAux re ec
        ECtl a x -> do
          gs <- getGS
          runCtl gs ec a x
        EAdd x -> do
          gs <- getGS
          runAdd gs ec x
        EBld r x -> do
          gs <- getGS
          build r cr team (x gs ec)
        EEnv x -> do
          gs <- getGS
          x gs ec
      ec = makeEC cr sr next
  where
    withRE f re ec = case team of
      Team0 -> f (Just (ec, re)) Nothing
      Team1 -> f Nothing (Just (ec, re))

makeEC :: CharRef -> SwordRef -> Maybe SwordRef -> EffectContext
makeEC cr sr next = EC
  { ecOwner = cr
  , ecThisSword = sr
  , ecNextSword = next
  }

performAttack :: (CharRef, CharRef) -> Maybe (EffectContext, [RichEffect]) -> Maybe (EffectContext, [RichEffect]) -> Game ()
performAttack (t0, t1) m0 m1 = do
  mapM_ (uncurry performRichEffect) $ zipMaybe (expandREs t0 m0) (expandREs t1 m1)
  where
    expandREs tgt m = case m of
      Nothing -> []
      Just (ctx, res) -> map ((,,) ctx tgt) res

performAux :: Maybe (EffectContext, [RichEffect]) -> Maybe (EffectContext, [RichEffect]) -> Game ()
performAux m0 m1 = mapM_ (uncurry performRichEffect) $ zipMaybe (expandREs m0) (expandREs m1)
  where
    expandREs m = case m of
      Nothing -> []
      Just (ctx, res) -> map ((,,) ctx (ecOwner ctx)) res

data MaybeDefendableAction
  = NotDefendable ActionChain
  | Defendable CharRef CharRef (CharRef -> ActionChain)

performRichEffect :: Maybe (EffectContext, CharRef, RichEffect) -> Maybe (EffectContext, CharRef, RichEffect) -> Game ()
performRichEffect m0 m1 = do
  gs <- getGS
  let friends x = gsTeam (csSide $ gsChars gs CR.! x) gs
  mapM_ (uncurry $ performRepeatedEffect) $ zipMaybe (expandAll friends m0) (expandAll friends m1)
  where
    expandAll friends m = case m of
      Nothing -> []
      Just (ec, obj, RE{ reRepeat = n, rePrim = prim }) -> case prim of
        TargetEnemy{ teDefendable = defn, teGen = gen, teAll = al } -> each_target al obj $ \tgt ->
          if defn then Defendable (ecOwner ec) tgt (flip gen ec) else NotDefendable $ gen tgt ec
        TargetAlly{ taGen = gen, taAll = al } -> each_target al (ecOwner ec) $ \tgt -> NotDefendable $ gen tgt ec
        TargetNone{ tnGen = gen } -> [(n, NotDefendable $ gen ec)]
        where
          each_target False obj f = [(n, f obj)]
          each_target True obj f = map ((,) n . f) $ friends obj

performRepeatedEffect :: Maybe (Int, MaybeDefendableAction) -> Maybe (Int, MaybeDefendableAction) -> Game ()
performRepeatedEffect m0 m1 = mapM_ (uncurry performWithDefence) $ zipMaybe (expandRepeat m0) (expandRepeat m1)
  where
    expandRepeat m = case m of
      Nothing -> []
      Just (n, mda) -> replicate n mda

performWithDefence :: Maybe MaybeDefendableAction -> Maybe MaybeDefendableAction -> Game ()
performWithDefence = loop
  where
    loop Nothing Nothing = return ()
    loop e0 e1 = do
      (a0, e0', descr0) <- handle e0
      (a1, e1', descr1) <- handle e1
      let descrs = descr0 ++ descr1
      performActionChain a0 a1
      when (not $ null descrs) $ mes $ intercalate "と" descrs ++ "が発動した"
      indented $ loop e0' e1'

    handle Nothing = return (Nothing, Nothing, [])
    handle (Just k) = case k of
      NotDefendable ac -> return (Just ac, Nothing, [])
      Defendable from to gen -> do
        cs <- getCS to
        case B.pop $ csStatusDefence cs of
          Nothing -> return (Just $ gen to, Nothing, [])
          Just (d, rest) -> do
            setCS to cs{ csStatusDefence = rest }
            return (Nothing, applyDef d, [show to ++ "の" ++ show d])
        where
          applyDef Reflect = Just $ Defendable to from gen
          applyDef Detect = Nothing

performActionChain :: Maybe ActionChain -> Maybe ActionChain -> Game ()
performActionChain a b = case (a, b) of
  (Just _, Just _) -> loop a b `equivalently` loop b a
  _ -> loop a b
  where
    loop Nothing Nothing = return ()
    loop x y = do
      gs <- getGS
      next0 <- runAC gs x
      next1 <- runAC gs y
      loop next0 next1
    runAC _ Nothing = return Nothing
    runAC gs (Just (AC f)) = liftM snd $ f gs

runCtl :: GameState -> EffectContext -> Bool -> (GameState -> EffectContext -> SwordState -> (Maybe SwordState, [String])) -> Game ()
runCtl gs ec@EC{ ecOwner = owner } scatter ctl = case targets of
  [] -> mes $ show owner ++ "には制御対象が存在しない"
  _ -> forM_ targets $ \(sr, ss) -> do
    let (ms, msgs) = ctl0 ss
    mapM_ mes msgs
    case ms of
      Just ss' -> modifyCS owner $ \k -> k{ csSwordState = M.insert sr ss' $ csSwordState k }
      Nothing -> modifyCS owner $ deleteSword sr
  where
    cs = gsChars gs CR.! owner
    targets
      | scatter = map (pair (csSwordState cs M.!)) $ csSwords cs
      | otherwise = maybeToList $ getECNext gs ec
    ctl0 = ctl gs ec

runAdd :: GameState -> EffectContext -> (GameState -> EffectContext -> Maybe SwordState) -> Game ()
runAdd gs ec add = whenJust (add gs ec) $ \ss -> do
  let ss' = ss{ ssUsed = False }
  mes $ show owner ++ "は" ++ show ss' ++ "を得た"
  addSword owner ss'
  return ()
  where
    owner = ecOwner ec

build :: Bool -> CharRef -> TeamRef -> Character -> Game ()
build restrict pcr team new = do
  turn <- getTurnCount
  fail_reason <- prepareToBuild turn
  case fail_reason of
    Just r -> mes $ r ++ "、" ++ show pcr ++ "はビルドに失敗した"
    Nothing -> do
      gs <- getGS
      let n = 2 * uniq gs + teamnum
      let name = filter (/='/') (chName new) ++ "/" ++ show n
      let cr = CR n (Just name)
      let br = if restrict then Just turn else Nothing
      let state = (initialCS CI{ ciDispName = name, ciSide = team, ciBuildRestriction = br, ciOriginal = new }) { csAp = chSpd new }
      mes $ name ++ "がビルドされた"
      mes $ show state
      setGS $ add team cr $ setUniq (uniq gs + 1) $ gs { gsChars = CR.insert cr state $ gsChars gs }
      when (csHp state <= 0) $ kill cr
  where
    add Team0 cr gs = gs{ gsTeam0 = cr : gsTeam0 gs }
    add Team1 cr gs = gs{ gsTeam1 = cr : gsTeam1 gs }

    uniq = teamcase gsUniqueSupply0 gsUniqueSupply1 team

    setUniq n = case team of
      Team0 -> \gs -> gs{ gsUniqueSupply0 = n }
      Team1 -> \gs -> gs{ gsUniqueSupply1 = n }

    teamnum = teamcase 0 1 team

    prepareToBuild turn = do
      cs <- getCS pcr
      case csStatusBad cs of
        SB{ sbAbortion = 0 }
          | Just turn == ciBuildRestriction (csInfo cs) -> return $ Just "生まれたてのため"
          | otherwise -> return Nothing
        sb -> do
          setCS pcr $ cs{ csStatusBad = sb{ sbAbortion = sbAbortion sb - 1 } }
          return $ Just "流産が発動したため"

cleanup :: Game ()
cleanup = do
  transformAllPresent $ \cr cs -> (cs{ csSwordState = M.map unuse $ csSwordState cs }, [])
  mes ""
  where
    unuse ss = ss{ ssUsed = False }

repeatedly :: (Monad m) => m Bool -> m ()
repeatedly a = do
  c <- a
  case c of
    True -> repeatedly a
    False -> return ()

checkpoint :: Game ()
checkpoint = eliminateDead >> checkGameEnd

eliminateDead :: Game ()
eliminateDead = do
  gs <- getGS
  when (gsSomeoneMayBeDead gs) $ do
    let (e0, new0) = elim gs $ gsTeam0 gs
    let (e1, new1) = elim gs $ gsTeam1 gs
    let e = e0 ++ e1
    when (not $ null e) $ mes $ intercalate "、" (map show e) ++ "は退場した"
    setGS $ gs{ gsTeam0 = new0, gsTeam1 = new1, gsSomeoneMayBeDead = False }
  where
    elim gs team = partition f team
      where
        cs = gsChars gs
        f c = csDead (cs CR.! c)

checkGameEnd :: Game ()
checkGameEnd = do
  gs <- getGS
  whenJust (cond gs) $ \reason -> do
    mes ""
    mes $ reason ++ "のため戦闘終了"
    terminate
  where
    cond gs
      | gsTurnCount gs > 20 = Just "ターンオーバー"
      | any null [gsTeam0 gs, gsTeam1 gs] = Just "全滅"
      | otherwise = Nothing

terminate :: Game a
terminate = do
  gs <- getGS
  let
    (reasons, result) = compareScore (score gs Team0) (score gs Team1)
    winner = case result of
      LT -> Just Team1
      GT -> Just Team0
      EQ -> Nothing
  mes $ intercalate "、" reasons ++ "により、" ++ show_result gs winner
  finish $ GR{ grWinner = winner }
  where
    score gs team
      | null live = (False, push members, hp members)
      | otherwise = (True, push live, hp live)
      where
        members = filter ((==team) . csSide) $ CR.elems $ gsChars gs
        live = filter (not . csDead) members
        push = sum . map (soPush . csStatusOther)
        hp = sum . map csHp

    compareScore (live0, p0, h0) (live1, p1, h1) = case compare live0 live1 of
      EQ -> addMsg msg $ comparePush (p0, h0) (p1, h1)
        where
          msg = if live0 then "両パーティ生存" else "両者全滅"
      r -> (["相手側のみ全滅"], r)
    comparePush (p0, h0) (p1, h1) = case compare p0 p1 of
      EQ
        | p0 == 0 -> compareHp h0 h1
        | otherwise -> addMsg msg $ compareHp h0 h1
        where
          msg = "両者プッシュ" ++ show p0
      r -> (["プッシュ" ++ show p0 ++ "対" ++ show p1], r)
    compareHp h0 h1 = (["HP" ++ show h0 ++ "対" ++ show h1], compare h0 h1)

    addMsg msg (m, p) = (msg:m, p)

    show_result gs k = case k of
      Just team -> teamname ++ "が勝利した"
        where
          teamname = ciDispName $ csInfo $ gsChars gs CR.! (teamcase primary0 primary1 team)
      Nothing -> "引き分けとなった"

transformAllPresent :: (CharRef -> CharacterState -> (CharacterState, [String])) -> Game ()
transformAllPresent f = do
  gs <- getGS
  mapM_ trans $ gsTeam0 gs ++ gsTeam1 gs
  where
    trans c = do
      cs <- getCS c
      let (cs', mss) = f c cs
      mapM_ mes mss
      setCS c cs'

ssEffect :: SwordState -> Effect
ssEffect SS{ ssSword = sw, ssAll = sall, ssPoison = poison, ssStun = stun, ssDestruction = destr }
  | not sall && poison == 0 && stun == 0 && destr == 0 = swEffect sw -- shortcut for efficiency
  | otherwise = case swEffect sw of
    EAtk re -> EAtk $ map fixRE re
    EAux re -> EAux $ map fixRE re
    e -> e
  where
    fixRE re = re{ rePrim = fixPE $ rePrim re }
    fixPE pe = case pe of
      TargetEnemy{ teAll = al, teGen = gen } ->
        pe{ teAll = al || sall, teGen = \tgt ec -> fixAC $ gen tgt ec }
      TargetAlly{ taAll = al, taGen = gen } ->
        TargetAlly{ taAll = al || sall, taGen = \tgt ec -> fixAC $ gen tgt ec }
      TargetNone{ tnGen = gen } ->
        TargetNone{ tnGen = \ec -> fixAC $ gen ec }
    fixAC =
      add poison putPoison damageDealingPhase .
      add stun putStun damageDealingPhase .
      add destr (retarget destroySword) destructionPhase
    add 0 _ _ = id
    add count action phase = addSideEffect phase $ \gs tgt ->
      replicateM_ count $ action gs tgt
    retarget action _gs tgt = do
      gs <- getGS
      action gs tgt

ssAvailableTurn :: SwordState -> Int
ssAvailableTurn ss = swTurn (ssSword ss) - ssEarly ss

ssReactsToDanger :: SwordState -> Bool
ssReactsToDanger ss = swReactsToDanger (ssSword ss) || ssForcedReactsToDanger ss

ssReactsToIllness :: SwordState -> Bool
ssReactsToIllness ss = swReactsToIllness (ssSword ss)

ssLimited :: SwordState -> Bool
ssLimited ss = swLimited (ssSword ss) || ssLimitAdded ss
