#!/usr/bin/env ruby

$KCODE = 'UTF-8'
$LOAD_PATH.unshift('./lib')

require 'cgi'
require 'open4'
require 'io/nonblock'
require 'htree'
require 'json'

class CGI
  def getlit(key)
    params = self.params[key]
    return '' unless params
    value = params[0]
    if value then value.dup else '' end
  end
  def get(n, key)
    getlit(key+n)
  end
end

@log = ''

def filter_with_timeout(input, timeout, pid, inn, out, err)
  writepos = 0
  waitend = Time.now + timeout

  pos = 0

  outs = ''
  errs = ''

  inn.nonblock = true
  out.nonblock = true
  err.nonblock = true

  opens = lambda{ |fds| fds.select{ |k| !k.closed? } }

  while (rem = waitend - Time.now) > 0 and r = select(opens[[out,err]], opens[[inn]], [], rem) do
    @log += "loop r=#{r.inspect} pos=#{pos} rem=#{waitend - Time.now}\n"
    if r[1][0]
      @log += "write\n"
      pos += inn.write(input[pos, 1024])
      inn.close if pos == input.size
    elsif r[0][0] == out
      @log += "read out\n"
      outs += out.read(16*1024) rescue out.close
    else
      @log += "read err\n"
      errs += err.read(16*1024) rescue err.close
    end
  end
  if inn.closed? && out.closed? && err.closed?
    # success
    [outs, errs]
  else
    @log += "timed out: #{waitend - Time.now}\n"
    Process.kill('TERM', pid)
    nil
  end
end

def make_pool(pool)
  case pool
  when '1a'
    '1a'
  when 'void'
    'void'
  else
    '100'
  end
end

def do_simulation(input, pool)
  pid, inn, out, err = Open4.popen4 "/usr/bin/env LANG=ja_JP.UTF-8 lib/rr-cgi #{pool} +RTS -M20m"
  out.gets # read out a ready signal
  filter_with_timeout input, 1, pid, inn, out, err
end

Page = Struct.new(:title, :head, :body)

def error_page(body)
  Page.new('エラー - m100sim', 'エラー', body)
end

def string_error_page(desc)
  error_page(expand{'<p _text="desc"/>'})
end

def expand(&block)
  template = '<t>' + block.call + '</t>'
  # HTree::TemplateCompiler is an undocumented class
  HTree(HTree::TemplateCompiler.new.expand_template(template, '', 'UTF-8', block)).root.children
end

def read_file(fname)
  open(fname) { |k| k.read }
end

def response_page(out)
  result = JSON.parse(out)
  notices = make_notices(result)
  prefix = notices.length > 0 ?  expand{ '<h2>お知らせ</h2><div _iter="notices.each//w"><div _tree=w /></div>' } : []
  Page.new('結果 - m100sim', '結果', prefix + HTree.compile_template(read_file('lib/result.xml')).result(result, result['chars'].length).children)
end

def make_notices(result)
  ret = []
  ignored = result['ignored']
  if ignored.length > 0
    msg = ignored.join("\n")
    ret += expand{'<p>以下の行には剣師が記述されているように見えますが、理解できなかったため無視しました。</p><pre _text=msg />'}
  end
  ret
end

def create_form(input, pool)
  HTree.compile_template(read_file('lib/form.xml')).form(input, pool).children
end

def output(input, pool, page)
  form = create_form(input, pool)
  sep = expand{ '<ul class=returntotop><li><a href=".">m100simトップに戻る</a></li></ul><h2 id="retry">再試行</h2>'}
  page1 = Page.new(page.title, page.head, page.body + sep + form);
  #page1 = page;
  body = ''
  HTree.expand_template('lib/page.xml', page1, body, 'UTF-8')

  puts "Content-Type: text/html; charset=UTF-8"
  puts "Content-Length: #{body.size}"
  puts
  print body
end

def make_page(input, pool)
  @log += "input=#{input}\npool=#{pool}\n"
  result = do_simulation(input, pool)
  if result
    out, err = result
    if err != ''
      if err =~ /^Heap exhausted;$/
        string_error_page 'シミュレーションが規定サイズのメモリ内で完了しませんでした。剣師の数を減らすか、ダウンロード版を試してみて下さい。'
      else
        raise "The simulator reported an error: #{err}"
      end
    else
      response_page out
    end
  else
    string_error_page 'シミュレーションが規定の時間内に完了しませんでした。剣師の数を減らすか、ダウンロード版を試してみて下さい。'
  end
end

def report_error(cgi, exn)
  msg = Iconv.conv('ISO-2022-JP', 'UTF-8', 
    "An error ocurred during an excution of m100sim-web.\nError:\n#{exn}\n\nLog:\n#{@log}\nInput:\n#{cgi.params.inspect}\n");
  IO.popen('/usr/sbin/sendmail -t', 'w') { |sm|
    sm.puts "To: web@kotha.net"
    sm.puts "From: m100sim-web error report <m100sim@kotha.net>"
    sm.puts "Subject: An error ocurred in m100sim-web"
    sm.puts "Content-Transfer-Encoding: 7bits"
    sm.puts "Content-Type: text/plain; charset=iso-2022-jp"
    sm.puts
    sm.puts msg
  }
  #File.open('errorlog', 'a') {|f| f.puts msg }
end

cgi = CGI.new
input = cgi.getlit('chars').gsub(/[\r\n]+/, "\n")
pool = make_pool cgi.getlit('pool')

page = 
  begin
    make_page(input, pool)
  rescue
    report_error(cgi, $!.to_s + "\n" + $@.join("\n"));
    string_error_page '想定外のエラーが発生したため処理を完遂できませんでした。申し訳ありません。管理者にバグ報告を送信しておきました。'
  end

output input, pool, page
