# ifndef M100SIM_INTERNAL_H
# define M100SIM_INTERNAL_H

# include <list>
# include <set>
# include <map>
# include <string>
# include <vector>
# include <iostream>

# include <stdarg.h>

# include "boost/optional.hpp"
# include "boost/function.hpp"
# include "boost/format.hpp"

# include "match.h"
# include "int.h"

namespace m100sim
{

using boost::optional;
using namespace std;

struct logger
{
  logger(FILE *fp, const char *term) : out(fp), level(0), term(term) {}
  void put(const boost::format &fmt) { put(fmt.str()); }
  void put(const string& s)
  {
    if(out)
    {
      for(int i = 0; i < level; i++)
        fputc(' ', out);
      fputs(s.c_str(), out);
      fputs(term, out);
    }
  }
  void putf(const char *fmt, ...)
  {
    if(out)
    {
      for(int i = 0; i < level; i++)
        fputc(' ', out);
      va_list ap;
      va_start(ap, fmt);
      vfprintf(out, fmt, ap);
      va_end(ap);
      fputs(term, out);
    }
  }
  void indent(int d)
  {
    level += 2*d;
  }
  FILE *start_direct()
  {
    if(out)
      for(int i = 0; i < level; i++)
        fputc(' ', out);
    return out;
  }
  void end_direct() { fputs(term, out); }
  int current_indent() const { return level / 2; }
  bool null() const { return !out; }
private:
  FILE *const out; /* may be null */
  int level;
  const char *const term;
};

struct indent_controller
{
  explicit indent_controller(logger &lg, int n = 0)
    : lg(lg), total(0)
  {
    add(n);
  }
  void add(int k)
  {
    lg.indent(k);
    total += k;
  }
  ~indent_controller()
  {
    lg.indent(-total);
  }
private:
  logger &lg;
  int total;
};

const int prio_env = 0;
const int prio_bld = 1;
const int prio_atk = 2;
const int prio_add = 3;
const int prio_ctl = 4;
const int prio_aux = 5;
const int prio_skp = 6;
const int n_priority = 7;

enum target_type
{
  target_enemy,
  target_ally,
  target_none
};

struct cstate;
struct sstate;
struct gstate;

typedef list<sstate>::iterator sref;

struct enchantments
{
  enchantments()
    : stun(0), poison(0), scatter(false), early(0), limited(false), reacts_to_danger(false), destruction(0) {}
  int stun;
  int poison;
  bool scatter;
  int early;
  bool limited;
  bool reacts_to_danger;
  int destruction;
};

struct effect_context
{
  sstate *this_sword;
  optional<sref> next_sword;
  cstate *owner;
  enchantments ench;
};

struct prim_effect_base
{
  /* represents "and" chain.
   * attacking effects are broken down into 5 components, which are glued together with "and"
   * "and" is similar to "そして", but is more fine-grained.
   * example:
   *   let X = (Attack - Target's Defence)
   *   and declare an X-point damage to Target (drive, guard, etc. are handled here)
   *   and cast the declared damage
   *   and if 1 or more points of damage were cast, let Y be the Target's first sword
   *   and move Y to the rightmost position of Self's sword list
   * similarly, non-attacking effects are decomposed into 2 parts:
   *  do nothing
   *  and increase Target's Defence by 1
   * each component corresponds to a call to `perform'.
   *
   * `perform' returns whether there is more thing to do
   */
  virtual bool perform(gstate &, const effect_context &, cstate *target, logger &) = 0;
  virtual ~prim_effect_base() {}
};

struct rich_effect
{
  rich_effect(target_type type, const boost::function<prim_effect_base *()> &gen, bool defendable = false, int n = 1, bool scatter = false)
    : count(n), scatter(scatter), type(type), gen(gen), defendable(defendable) {}
  int count;
  bool scatter;
  target_type type;
  boost::function<prim_effect_base *()> gen;
  bool defendable;
};

typedef vector<rich_effect> effect;

struct sinfo
{
  sinfo(const char *name, const char *sname, const effect &eff, int priority,
      int turn = 1, bool limited = false, optional<int> mp = boost::none, optional<int> period = boost::none,
      bool danger = false, bool illness = false)
    : name(name), shortname(sname), turn(turn), limited(limited), period(period), mp(mp), eff(eff), prio(priority)
    , reacts_to_danger(danger), reacts_to_illness(illness)
  {}
  const char *name;
  const char *shortname;
  int turn;
  bool limited;
  optional<int> period;
  optional<int> mp;
  effect eff;
  int prio;
  bool reacts_to_danger;
  bool reacts_to_illness;
};

inline const sinfo *sinfo_by_ref(m100sim_sword_ref k)
{
  return (const sinfo *)k;
}

struct sstate
{
  explicit sstate(m100sim_sword_ref k)
    : info(sinfo_by_ref(k)), used(false) {}
  const sinfo *info;
  bool used;
  enchantments ench;
};

enum side_id
{
  Left = 0,
  Right = 1
};

enum guard_status
{
  barrier,
  cusion,
  flip,
  shield,
  guard
};

enum defence_status
{
  reflection,
  detection
};

enum revival_status
{
  soul,
  life
};

struct status_bad
{
  status_bad() : stun(0), sleep(0), bad_poison(0), poison(0), abortion(0) {}
  int stun;
  int sleep;
  int bad_poison;
  int poison;
  int abortion;
  bool empty() const
  {
    return !stun && !sleep && !bad_poison && !poison && !abortion;
  }
};

struct status_other
{
  status_other() : push(0), heavens_door(false), quick(0), slow(0), drive(0), provocation(false), heavy_push(0) {}
  int push;
  bool heavens_door;
  int quick;
  int slow;
  int drive;
  bool provocation;
  int heavy_push;
};

struct cinfo
{
  cinfo(const m100sim_character &orig, const string &name, side_id side, int number, optional<int> restriction)
    : name(name), side(side), number(number), build_restriction(restriction), original(orig) {}
  string name;
  side_id side;
  int number; /* number within a team */
  optional<int> build_restriction;
  m100sim_character original;
};

struct cstate
{
  explicit cstate(const cinfo &inf)
    : max_hp(inf.original.hp)
    , hp(inf.original.hp)
    , atk(inf.original.atk)
    , def(inf.original.def)
    , spd(inf.original.spd)
    , mp(0)
    , ap(0)
    , reserved_ap(0)
    , dead(false)
    , info(inf)
  {
    for(int i = 0; i < inf.original.n_swords; i++)
      swords.push_back(sstate(inf.original.swords[i]));
  }
  int_t max_hp;
  int_t hp;
  int_t atk;
  int_t def;
  int_t spd;
  int_t mp;
  int_t ap;
  int_t reserved_ap;
  bool dead;
  list<sstate> swords;
  status_bad bad;
  status_other other;
  multiset<guard_status> guard;
  multiset<defence_status> defence;
  multiset<revival_status> revival;
  cinfo info;
};

struct character_order
{
  bool operator()(cstate *x, cstate *y) const
  {
    const cinfo &xi = x->info, &yi = y->info;
    if(xi.side == yi.side) return xi.number > yi.number; /* character of larger number is considered smaller */
    return xi.side < yi.side;
  }
};

typedef set<cstate *, character_order> cset;

struct tstate
{
  tstate(const m100sim_character &prim, const string &name, side_id side)
    : hp_removed(0), push_removed(0), unique_source(0), side(side), primary(&prim), name(name)
  {}
  list<cstate> chs;
  int_t hp_removed;
  int push_removed;
  int unique_source;
  side_id side;
  const m100sim_character *primary;
  string name;
};

struct quickness
{
  quickness(int nq, intparam spd) : nquick(nq), spd(spd) {}
  explicit quickness(const cstate *cs) : nquick(cs->other.quick - cs->other.slow), spd(cs->spd) {}
  int nquick; /* may be negative */
  int_t spd;
};

inline bool operator>(const quickness &x, const quickness &y)
{
  if(x.nquick == y.nquick) return x.spd > y.spd;
  return x.nquick > y.nquick;
}

inline bool operator<(const quickness &x, const quickness &y)
{
  if(x.nquick == y.nquick) return x.spd < y.spd;
  return x.nquick < y.nquick;
}

struct gstate
{
  gstate(const m100sim_character &lc, const m100sim_character &rc)
    : turn(0), action_count(0), someone_may_be_dead(false)
  {
    string lname = lc.name ? lc.name : "";
    string rname = rc.name ? rc.name : "";
    if(!strcmp(lc.name, rc.name))
    {
      lname += "/0";
      rname += "/1";
    }
    teams.push_back(tstate(lc, lname, Left));
    teams.push_back(tstate(rc, rname, Right));
    add_character(cinfo(lc, lname, Left, teams[Left].unique_source++, boost::none));
    add_character(cinfo(rc, rname, Right, teams[Right].unique_source++, boost::none));
    move_next = action_order.end();
  }
  int turn;
  int action_count; /* for ensuring no overflow */
  vector<tstate> teams;
  bool someone_may_be_dead; 
  /* removed swords are temporarily stored here, so that 
   * references to them won't be invalidated too early.
   */
  list<sstate> sword_trash;
  typedef map<quickness, cset, greater<quickness> > action_order_type;
  action_order_type action_order;
  action_order_type::iterator move_next;

  cstate *add_character(const cinfo &c);
};

/* common operations */

void revive(gstate &gs, cstate *tgt, logger &lg);
void really_kill(gstate &gs, cstate *tgt, logger &lg);
bool invoke_revival(gstate &gs, cstate *tgt, logger &lg);
void kill(gstate &gs, cstate *tgt, logger &lg);
void set_hp(gstate &gs, cstate *tgt, intparam val, logger &lg);
void damage(gstate &gs, const char *reason, cstate *tgt, intparam amount, logger &lg);
inline side_id opponent(side_id k) { return (side_id)!k; }
const char *name(guard_status s);
const char *name(revival_status s);
const char *name(defence_status s);
void increment_turn_count(gstate &gs, logger &lg);
inline void remove_sword(gstate &gs, cstate *owner, sref s) { gs.sword_trash.splice(gs.sword_trash.end(), owner->swords, s); }
void register_to_action_order(gstate &gs, cstate *c);
void unregister_from_action_order(gstate &gs, cstate *c);
void print_cstate(FILE *out, const cstate &cs);
void print_sstate(FILE *out, const sstate &ss, bool show_usedness = true);

# define DEFINE_OUTPUT_BY_NAME(type) inline ostream &operator<<(ostream &out, type t) { return out << name(t); }
DEFINE_OUTPUT_BY_NAME(guard_status)
DEFINE_OUTPUT_BY_NAME(revival_status)
DEFINE_OUTPUT_BY_NAME(defence_status)
# undef DEFINE_OUTPUT_BY_NAME

} /* namespace m100sim */

# endif
