{-# OPTIONS_GHC -cpp -optP-P -fglasgow-exts #-}
-- #hide
{- |
   Module     : System.IO.Fd
   Copyright  : Copyright (C) 2006 Takano Akio
                Copyright (C) 2006 Bulat Ziganshin
                (c) The University of Glasgow, 1992-2002

   License    : BSD3

   Maintainer : Takano Akio <aljee@hyper.cx>
   Stability  : experimental
   Portability: Hugs/GHC

Raw POSIX files support
-}

module System.IO.Fd
  ( openRawTextFd
  , openRawBinaryFd
  , openTempFd
  , fdOpen
  , fdClose
  , fdSetFlags
  , fdGetBuf
  , fdGetBufNonBlocking
  , fdPutBuf
  , fdPutBufNonBlocking
  , fdRecv
  , fdRecvNonBlocking
  , fdSend
  , fdSendNonBlocking
  , fdTell
  , fdSeek
  , fdFileSize
  , fdSetFileSize
  , fdWaitForInput
  , fdReady
  , fdSync
  , fdIsTerminal
  , fdSetEcho
  , fdGetEcho
  , System.Posix.Types.Fd(..)
  )
  where

import Prelude hiding (catch)
import Control.Concurrent
import Control.Monad
import Control.OldException
import Data.Bits
import Data.Int
import Data.Word
import Foreign.C.Types
import Foreign.C.Error
import Foreign.C.String
import Foreign.Marshal.Alloc
import Foreign.Ptr
import System.IO
import System.IO.Error hiding (catch)
import System.Posix.Types

# if defined __GLASGOW_HASKELL__ && defined mingw32_HOST_OS
import GHC.Conc(asyncRead, asyncWrite, asyncDoProc)
# endif

type CWFilePath   = CString   -- filename in C land
type CWFileOffset = COff      -- filesize or filepos in C land
type FileSize     = Integer   -- filesize or filepos in Haskell land
withCWFilePath = withCString  -- FilePath->CWFilePath conversion
peekCWFilePath = peekCString  -- CWFilePath->FilePath conversion

-- ---------------------------------------------------------------------------
-- Opening and Closing Files

-- | Computation 'openRawTextFd' @file mode@ allocates and returns a new, open
-- handle to manage the file @file@.  It manages input if @mode@
-- is 'ReadMode', output if @mode@ is 'WriteMode' or 'AppendMode',
-- and both input and output if mode is 'ReadWriteMode'.
--
-- If the file does not exist and it is opened for output, it should be
-- created as a new file.  If @mode@ is 'WriteMode' and the file
-- already exists, then it should be truncated to zero length.
-- Some operating systems delete empty files, so there is no guarantee
-- that the file will exist following an 'openFile' with @mode@
-- 'WriteMode' unless it is subsequently written to successfully.
-- The handle is positioned at the end of the file if @mode@ is
-- 'AppendMode', and otherwise at the beginning (in which case its
-- internal position is 0).
-- The initial buffer mode is implementation-dependent.
--
-- This operation may fail with:
--
--  * 'isAlreadyInUseError' if the file is already open and cannot be reopened;
--
--  * 'isDoesNotExistError' if the file does not exist; or
--
--  * 'isPermissionError' if the user does not have permission to open the file.
--
-- Note: if you will be working with files containing binary data, you'll want to
-- be using 'openRawBinaryFd'.
openRawTextFd :: FilePath -> IOMode -> IO Fd
openRawTextFd fp im =
  modifyIOError
    (\e -> annotateIOError e "openRawTextFd" Nothing (Just fp))
    (openRawFd' fp im False)

-- | Like 'openRawTextFd', but open the file in binary mode.
-- On Windows, reading a file in text mode (which is the default)
-- will translate CRLF to LF, and writing will translate LF to CRLF.
-- This is usually what you want with text files.  With binary files
-- this is undesirable; also, as usual under Microsoft operating systems,
-- text mode treats control-Z as EOF.  Binary mode turns off all special
-- treatment of end-of-line and end-of-file characters.
-- (See also 'hSetBinaryMode'.)

openRawBinaryFd :: FilePath -> IOMode -> IO Fd
openRawBinaryFd fp m =
  modifyIOError
    (\e -> annotateIOError e "openRawBinaryFd" Nothing (Just fp))
    (openRawFd' fp m True)

openRawFd' :: FilePath -> IOMode -> Bool -> IO Fd
openRawFd' filepath mode binary =
    let
      oflags1 = case mode of
                  ReadMode      -> read_flags
                  WriteMode     -> write_flags .|. o_TRUNC
                  ReadWriteMode -> rw_flags
                  AppendMode    -> append_flags

      binary_flags
          | binary    = o_BINARY
          | otherwise = 0

      oflags = oflags1 .|. binary_flags

    in do

    -- the old implementation had a complicated series of three opens,
    -- which is perhaps because we have to be careful not to open
    -- directories.  However, the man pages I've read say that open()
    -- always returns EISDIR if the file is a directory and was opened
    -- for writing, so I think we're ok with a single open() here...
    fd <- fdOpen filepath oflags 0o666

    return fd


openTempFd :: FilePath -> String -> Bool -> IO (FilePath, Fd)
openTempFd tmp_dir template binary = do
  pid <- c_getpid
  findTempName pid
  where
    (prefix,suffix) = break (=='.') template

    oflags1 = rw_flags .|. o_EXCL

    binary_flags
      | binary    = o_BINARY
      | otherwise = 0

    oflags = oflags1 .|. binary_flags

    findTempName x = do
      fd <- withCString filepath $ \ f ->
              c_open f oflags 0o666
      if fd < 0 
       then do
         errno <- getErrno
         if errno == eEXIST
           then findTempName (x+1)
           else ioError (errnoToIOError "System.IO.Fd.openTempFd" errno Nothing (Just tmp_dir))
       else return (filepath, fromIntegral fd)
      where
        filename        = prefix ++ show x ++ suffix
        filepath        = tmp_dir `joinFileName` filename
        dir `joinFileName` fname = dir ++ "/" ++ fname

-- Flags for fdOpen
# if defined __GLASGOW_HASKELL__ && !defined mingw32_HOST_OS
std_flags    = o_NONBLOCK   .|. o_NOCTTY
# else
std_flags    = o_NOCTTY
# endif
output_flags = std_flags    .|. o_CREAT
read_flags   = std_flags    .|. o_RDONLY
write_flags  = output_flags .|. o_WRONLY
rw_flags     = output_flags .|. o_RDWR
append_flags = write_flags  .|. o_APPEND

-- -----------------------------------------------------------------------------
-- Low-level functions

fdOpen :: String -> CInt -> CMode -> IO Fd
fdOpen name access mode =
  toFd $
    modifyIOError (`ioeSetFileName` name) $
      withCWFilePath name $ \ p_name ->
        throwErrnoIfMinus1Retry "fdOpen" $
          c_open p_name access mode

fdClose :: Fd -> IO ()
fdClose fd =
  throwErrnoIfMinus1Retry_ "fdClose" $
    c_close (unFd fd)

fdSetFlags :: Fd -> IO ()
# if defined __GLASGOW_HASKELL__ && !defined mingw32_HOST_OS
fdSetFlags fd = do
  flags <- 
    throwErrnoIfMinus1Retry "fdSetFlags" $
      c_get_status_flags (unFd fd)
  let 
    flags' = 
      flags .|. o_NONBLOCK
  throwErrnoIfMinus1Retry_ "fdSetFlags" $
    c_set_status_flags (unFd fd) flags'

foreign import ccall unsafe "fdio.h fdio_get_status_flags" c_get_status_flags :: CInt -> IO CInt
foreign import ccall unsafe "fdio.h fdio_set_status_flags" c_set_status_flags :: CInt -> CInt -> IO CInt

# else
fdSetFlags fd = return ()
# endif


fdGetBuf, fdPutBuf, fdRecv, fdSend :: Fd -> Ptr a -> Int -> IO Int
fdGetBufNonBlocking, fdPutBufNonBlocking, fdRecvNonBlocking, fdSendNonBlocking :: Fd -> Ptr a -> Int -> IO (Maybe Int)

# ifdef __GLASGOW_HASKELL__
#  ifdef mingw32_HOST_OS

fdGetBuf fd buf size
  | threaded = 
    fromIntegral `liftM` 
      (throwErrnoIfMinus1Retry "fdGetBuf" $
        c_safe_read (unFd fd) (castPtr buf) (fromIntegral size))
  | otherwise = fdReadA "fdGetBuf" fd False size buf

fdGetBufNonBlocking fd buf size = do
  ready <- fdReady fd False
    `catch` \e -> return True
  if ready
    then liftM Just $ fdGetBuf fd buf size
    else return Nothing

fdPutBuf fd buf size
  | threaded = 
    fromIntegral `liftM`
      (throwErrnoIfMinus1Retry "fdPutBuf" $
        c_safe_write (unFd fd) (castPtr buf) (fromIntegral size))
  | otherwise = fdWriteA "fdPutBuf" fd False size buf 

-- We cannot wait for non-socket output fds on Win32.
fdPutBufNonBlocking fd buf size = liftM Just $ fdPutBuf fd buf size

fdRecv fd buf size
  | threaded = 
    fromIntegral `liftM`
      (throwErrnoIfMinus1Retry "fdRecv" $
        c_safe_recv (unFd fd) (castPtr buf) (fromIntegral size) 0)
  | otherwise = fdReadA "fdRecv" fd True size buf

fdRecvNonBlocking fd buf size = do
  ready <- fdReady fd True
    `catch` \e -> return True
  if ready
    then liftM Just $ fdRecv fd buf size
    else return Nothing

fdSend fd buf size
  | threaded = 
    fromIntegral `liftM`
      (throwErrnoIfMinus1Retry "fdSend" $
        c_safe_send (unFd fd) (castPtr buf) (fromIntegral size) 0)
  | otherwise = fdWriteA "fdSend" fd True size buf

fdSendNonBlocking fd buf size = do
  ready <- liftM (/=0) $
    throwErrnoIfMinus1 "fdSendNonBlocking" $ c_winsock_output_ready (fromIntegral fd)
  if ready
    then liftM Just $ fdSend fd buf size
    else return Nothing

fdReadA :: String -> Fd -> Bool -> Int -> Ptr a -> IO Int
fdReadA loc fd isSock size buf = do
  (l, rc) <- asyncRead (fromIntegral $ unFd fd) (if isSock then 1 else 0) size (castPtr buf)
  if l == (-1)
    then ioError (errnoToIOError loc (Errno (fromIntegral rc)) Nothing Nothing)
    else return l

fdWriteA :: String -> Fd -> Bool -> Int -> Ptr a -> IO Int
fdWriteA loc fd isSock size buf = do
  (l, rc) <- asyncWrite (fromIntegral $ unFd fd) (if isSock then 1 else 0) size (castPtr buf)
  if l == (-1)
    then ioError (errnoToIOError loc (Errno (fromIntegral rc)) Nothing Nothing)
    else return l

foreign import ccall safe "unistd.h read" c_safe_read :: CInt -> Ptr () -> CSize -> IO CSize
foreign import ccall safe "unistd.h write" c_safe_write :: CInt -> Ptr () -> CSize -> IO CSize
foreign import stdcall safe "winsock.h recv" c_safe_recv :: CInt -> Ptr () -> CSize -> CInt -> IO CSize
foreign import stdcall safe "winsock.h send" c_safe_send :: CInt -> Ptr () -> CSize -> CInt -> IO CSize
foreign import ccall unsafe "fdio.h fdio_winsock_output_ready" c_winsock_output_ready :: CInt -> IO CInt

#  else /* mingw32_HOST_OS */

fdGetBuf fd buf size =
  fromIntegral `liftM`
    (throwErrnoIfMinus1RetryMayBlock "fdGetBuf"
      (c_read (unFd fd) (castPtr buf) (fromIntegral size))
      (threadWaitRead fd))

fdGetBufNonBlocking fd buf size = do
  r <- c_read (unFd fd) (castPtr buf) (fromIntegral size)
  if r == (-1)
    then do
      errno <- getErrno
      if errno == eAGAIN || errno == eWOULDBLOCK
        then return Nothing
        else throwErrno "fdGetBufNonBlocking"
    else return $ Just $ fromIntegral r

fdPutBuf fd buf size =
  fromIntegral `liftM` 
    throwErrnoIfMinus1RetryMayBlock "fdPutBuf"
      (c_write (unFd fd) (castPtr buf) (fromIntegral size))
      (threadWaitWrite fd)

fdPutBufNonBlocking fd buf size = do
  r <- c_write (unFd fd) (castPtr buf) (fromIntegral size)
  if r == (-1)
    then do
      errno <- getErrno
      if errno == eAGAIN || errno == eWOULDBLOCK
        then return Nothing
        else throwErrno "fdPutBufNonBlocking"
    else return $ Just $ fromIntegral r

fdRecv = fdGetBuf
fdRecvNonBlocking = fdGetBufNonBlocking
fdSend = fdPutBuf
fdSendNonBlocking = fdPutBufNonBlocking

foreign import ccall unsafe "unistd.h read" c_read :: CInt -> Ptr () -> CSize -> IO CSize
foreign import ccall unsafe "unistd.h write" c_write :: CInt -> Ptr () -> CSize -> IO CSize

#  endif /* !mingw32_HOST_OS */

# else /* __GLASGOW_HASKELL__ */

-- Non-blocking operations are not supported for compilers other than
-- GHC.

fdGetBuf fd buf size =
  fromIntegral `liftM`
    (throwErrnoIfMinus1Retry "fdGetBuf" $
      c_read (unFd fd) (castPtr buf) (fromIntegral size))

fdGetBufNonBlocking fd buf size = liftM Just $ fdGetBuf fd buf size

fdPutBuf fd buf size =
  fromIntegral `liftM` 
    throwErrnoIfMinus1Retry "fdPutBuf" 
      (c_write (unFd fd) (castPtr buf) (fromIntegral size))

fdPutBufNonBlocking fd buf size = liftM Just $ fdPutBuf fd buf size

fdRecv fd buf size =
  fromIntegral `liftM`
    throwErrnoIfMinus1Retry "fdRecv"
      (c_recv (unFd fd) (castPtr buf) (fromIntegral size) 0)

fdRecvNonBlocking fd buf size = liftM Just $ fdRecv fd buf size

fdSend fd buf size =
  fromIntegral `liftM` 
    throwErrnoIfMinus1Retry "fdSend" 
      (c_send (unFd fd) (castPtr buf) (fromIntegral size) 0)

fdSendNonBlocking fd buf size = liftM Just $ fdSend fd buf size

foreign import ccall unsafe "unistd.h read" c_read :: CInt -> Ptr () -> CSize -> IO CSize
foreign import ccall unsafe "unistd.h write" c_write :: CInt -> Ptr () -> CSize -> IO CSize
# ifdef mingw32_HOST_OS
foreign import stdcall unsafe "winsock.h recv" c_recv :: CInt -> Ptr () -> CSize -> CInt -> IO CSize
foreign import stdcall unsafe "winsock.h send" c_send :: CInt -> Ptr () -> CSize -> CInt -> IO CSize
# else
foreign import ccall unsafe "sys/socket.h recv" c_recv :: CInt -> Ptr () -> CSize -> CInt -> IO CSize
foreign import ccall unsafe "sys/socket.h send" c_send :: CInt -> Ptr () -> CSize -> CInt -> IO CSize
# endif

# endif /* !__GLASGOW_HASKELL__ */

fdTell :: Fd -> IO FileSize
fdTell fd =
  fromIntegral `liftM`
    throwErrnoIfMinus1Retry "fdTell"
      (c_lseek (unFd fd) sEEK_CUR 0)

fdSeek :: Fd -> SeekMode -> FileSize -> IO ()
fdSeek fd mode offset =
  throwErrnoIfMinus1Retry_ "fdSeek" $
    c_lseek (unFd fd) whence (fromIntegral offset) 
  where whence = case mode of
                   AbsoluteSeek -> sEEK_SET
                   RelativeSeek -> sEEK_CUR
                   SeekFromEnd  -> sEEK_END

fdFileSize :: Fd -> IO FileSize
fdFileSize fd = do
  fs <- c_filesize (unFd fd)
  e <- getErrno
  if e /= eOK
    then throwErrno "fdFileSize"
    else return $ fromIntegral fs

fdSetFileSize :: Fd -> FileSize -> IO ()
fdSetFileSize fd size =
  throwErrnoIfMinus1Retry_ "fdSetFileSize" $
    c_ftruncate (unFd fd) (fromIntegral size)

fdWaitForInput :: Fd -> Bool -> Int -> IO Bool
# ifdef __GLASGOW_HASKELL__
#  ifdef mingw32_HOST_OS
fdWaitForInput fd isSock msec = liftM (/=0) $ fdWaitForInputP (unFd fd) (fromIntegral msec) (if isSock then 1 else 0)

fdWaitForInputP :: CInt -> CInt -> CInt -> IO CInt
fdWaitForInputP fd msec isSock
  | threaded = throwErrnoIfMinus1 "fdWaitForInput" (c_safe_ready fd msec isSock)
  | otherwise = do
      d <- throwErrnoIfNull "fdWaitForInput" (c_new_wait_d fd msec isSock)
      rc <- asyncDoProc c_wait_proc d
      r <- c_wait_retval d
      c_delete_wait_d d
      if r == (-1)
        then ioError (errnoToIOError "fdWaitForInput" (Errno (fromIntegral rc)) Nothing Nothing)
        else return r

foreign import ccall safe "fdio.h fdio_ready" c_safe_ready :: CInt -> CInt -> CInt -> IO CInt
foreign import ccall unsafe "fdio.h fdio_new_wait_d" c_new_wait_d :: CInt -> CInt -> CInt -> IO (Ptr ())
foreign import ccall unsafe "fdio.h &fdio_run_wait_d" c_wait_proc :: FunPtr (Ptr () -> IO Int)
foreign import ccall unsafe "fdio.h fdio_wait_retval" c_wait_retval :: Ptr () -> IO CInt
foreign import ccall unsafe "fdio.h fdio_delete_wait_d" c_delete_wait_d :: Ptr () -> IO ()

#  else /* mingw32_HOST_OS */

fdWaitForInput fd isSock msec
  | msec < 0  = threadWaitRead fd >> return True
  | msec > 0  = fdWaitForInputTimeout fd msec
  | otherwise = fdWaitForInputC fd isSock 0

fdWaitForInputTimeout :: Fd -> Int -> IO Bool
fdWaitForInputTimeout fd msec = 
  (throwIfTimeout >> return True)
    `catch`
  (\e -> if e == timeoutError
    then return False
    else throwIO e)
  where
    throwIfTimeout = block $ do
      this <- myThreadId
      child <- forkIO (subthread this)
      (unblock $ threadWaitRead fd)
        `finally` killThread child
    subthread parent = do
      threadDelay (1000 * msec)
      throwTo parent timeoutError 
    timeoutError = IOException $ userError "fWFIT"

#  endif /* !mingw32_HOST_OS */
# else

fdWaitForInput = fdWaitForInputC

# endif /* !__GLASGOW_HASKELL__ */

fdWaitForInputC :: Fd -> Bool -> Int -> IO Bool
fdWaitForInputC fd isSock msec =
  liftM (/=0) $ throwErrnoIfMinus1 "fdReady" (c_ready (unFd fd) (fromIntegral msec) (if isSock then 1 else 0))

fdReady :: Fd -> Bool -> IO Bool
fdReady fd isSock = fdWaitForInputC fd isSock 0

fdIsTerminal :: Fd -> IO Bool
fdIsTerminal fd = liftM (/=0) $ c_isatty (unFd fd)

fdSetEcho :: Fd -> Bool -> IO ()
fdSetEcho fd set = do
  tty <- fdIsTerminal fd
  when (not tty) $ ioError $ userError "System.IO.SSC.fdSetEcho: not a terminal device"
  throwErrnoIfMinus1_ "fdSetEcho" $
    c_setecho (fromIntegral fd) (if set then 1 else 0)

fdGetEcho :: Fd -> IO Bool
fdGetEcho fd = do
  tty <- fdIsTerminal fd
  when (not tty) $ ioError $ userError "System.IO.SSC.fdGetEcho: not a terminal device"
  liftM (/=0) $ throwErrnoIfMinus1 "fdGetEcho" $
    c_getecho (fromIntegral fd)

fdSync :: Fd -> IO ()
fdSync fd = c_sync (unFd fd)

{-open/close/truncate/dup

  new_fd <- throwErrnoIfMinus1 "dupHandle" $
                c_dup (fromIntegral (haFd h_))
  new_fd <- throwErrnoIfMinus1 "dupHandleTo" $
                c_dup2 (fromIntegral (haFd h_)) (fromIntegral (haFd hto_))
-}


foreign import ccall unsafe "fdio.h fdio_seek_cur" sEEK_CUR :: CInt
foreign import ccall unsafe "fdio.h fdio_seek_set" sEEK_SET :: CInt
foreign import ccall unsafe "fdio.h fdio_seek_end" sEEK_END :: CInt
foreign import ccall unsafe "fdio.h fdio_o_trunc" o_TRUNC :: CInt
foreign import ccall unsafe "fdio.h fdio_o_binary" o_BINARY :: CInt
foreign import ccall unsafe "fdio.h fdio_o_nonblock" o_NONBLOCK :: CInt
foreign import ccall unsafe "fdio.h fdio_o_noctty" o_NOCTTY :: CInt
foreign import ccall unsafe "fdio.h fdio_o_creat" o_CREAT :: CInt
foreign import ccall unsafe "fdio.h fdio_o_rdonly" o_RDONLY :: CInt
foreign import ccall unsafe "fdio.h fdio_o_wronly" o_WRONLY :: CInt
foreign import ccall unsafe "fdio.h fdio_o_rdwr" o_RDWR :: CInt
foreign import ccall unsafe "fdio.h fdio_o_append" o_APPEND :: CInt
foreign import ccall unsafe "fdio.h fdio_o_excl" o_EXCL :: CInt

foreign import ccall unsafe "unistd.h open" c_open :: CString -> CInt -> CMode -> IO CInt
foreign import ccall unsafe "unistd.h close" c_close :: CInt -> IO CInt
foreign import ccall unsafe "unistd.h isatty" c_isatty :: CInt -> IO CInt
foreign import ccall unsafe "fdio.h fdio_lseek" c_lseek :: CInt -> CInt -> COff -> IO COff
foreign import ccall unsafe "unistd.h ftruncate" c_ftruncate :: CInt -> COff -> IO CInt

foreign import ccall unsafe "fdio.h fdio_ready" c_ready :: CInt -> CInt -> CInt -> IO CInt
foreign import ccall unsafe "fdio.h fdio_filesize" c_filesize :: CInt -> IO COff
foreign import ccall unsafe "fdio.h fdio_setecho" c_setecho :: CInt -> CInt -> IO Int
foreign import ccall unsafe "fdio.h fdio_getecho" c_getecho :: CInt -> IO CInt
foreign import ccall unsafe "fdio.h fdio_sync" c_sync :: CInt -> IO ()
foreign import ccall unsafe "fdio.h fdio_getpid" c_getpid :: IO CUInt

# ifdef __GLASGOW_HASKELL__
foreign import ccall "rtsSupportsBoundThreads" threaded :: Bool
# endif

i=fromIntegral
unFd (Fd x) = x
toFd :: (Monad m) => m CInt -> m Fd
toFd = liftM Fd

