{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Automatic-locking port transformer.
-}
module System.IO.SSC.AutoLock
  ( -- * Construction functions
    iAutoLock
  , oAutoLock
  , ioAutoLock
    -- * Transformer
  , AutoLock
  ) where

import Control.Concurrent.MVar
import System.IO.SSC.Classes
import System.IO.Error

data AutoLock a ma ba dest pos s = AL {-# UNPACK #-} !s {-# UNPACK #-} !(MVar Bool)

-- | Add automatic locking\/unlocking facility to any lockable port.
autoLock, iAutoLock, oAutoLock, ioAutoLock :: (Lockable s) => s -> IO (AutoLock a ma ba dest pos s)
autoLock s = do
  r <- newMVar False
  return $ AL s r
iAutoLock = autoLock
oAutoLock = autoLock
ioAutoLock = autoLock

finally :: IO a -> IO () -> IO a
a `finally` b = do
  r <- a `catch` \e -> b >> ioError e
  b
  return r

{-# INLINE withLock #-}
withLock :: (Lockable s) => s -> MVar Bool -> IO a -> IO a
withLock s r a = do
  got <- sTryLock s
  if got
    then a `finally` sUnsafeUnlock s
    else waitAndLock s r a

waitAndLock :: (Lockable s) => s -> MVar Bool -> IO a -> IO a
waitAndLock s r a = do
  locked <- takeMVar r
  r <- (if locked
    then a
    else do
      sLock s
      a `finally` sUnsafeUnlock s)
    `finally` putMVar r locked
  return r

instance (Lockable s, Port s) => Port (AutoLock a ma ba dest pos s) where
  rawClose (AL s r) = withLock s r $ rawClose s
  rawIsClosed (AL s r) = withLock s r $ rawIsClosed s
  rawFlush (AL s r) = withLock s r $ rawFlush s

instance (Lockable s, Destination dest s) => Destination dest (AutoLock a ma ba dest pos s) where
  rawDest (AL s r) = withLock s r $ rawDest s

instance (Lockable s) => Lockable (AutoLock a ma ba dest pos s) where
  sLock (AL s r) = loop
    where
      loop = do
        locked <- takeMVar r
        if locked
          then do
            putMVar r locked
            sLock s
            sUnsafeUnlock s
            loop
          else do
            sLock s
            putMVar r True
  sTryLock (AL s r) = do
    result <- tryTakeMVar r
    case result of
      Nothing -> return False
      Just True -> return False
      Just False -> do
        baseLocked <- sTryLock s
        if baseLocked
          then do
            putMVar r True
            return True
          else do
            putMVar r False
            return False
  sUnsafeUnlock (AL s r) = do
    takeMVar r
    sUnsafeUnlock s
    putMVar r False

instance (Lockable s, Discardable s) => Discardable (AutoLock a ma ba dest pos s) where
  rawDiscard (AL s r) = withLock s r $ rawDiscard s

instance (Lockable s, Seekable pos s) => Seekable pos (AutoLock a ma ba dest pos s) where
  rawGetPos (AL s r) = withLock s r $ rawGetPos s
  rawSetPos (AL s r) pos = withLock s r $ rawSetPos s pos

instance (Lockable s, RelativeSeekable pos s) => RelativeSeekable pos (AutoLock a ma ba dest pos s) where
  rawRelativeSeek (AL s r) pos = withLock s r $ rawRelativeSeek s pos

instance (Lockable s, SizedInput pos s) => SizedInput pos (AutoLock a ma ba dest pos s) where
  rawInputEnd (AL s r) = withLock s r $ rawInputEnd s
  rawSeekFromEnd (AL s r) pos = withLock s r $ rawSeekFromEnd s pos

instance (Lockable s, SizedOutput pos s) => SizedOutput pos (AutoLock a ma ba dest pos s) where
  rawOutputEnd (AL s r) = withLock s r $ rawOutputEnd s

instance (Lockable s, IStream a s) => IStream a (AutoLock a ma ba dest pos s) where
  rawGet (AL s r) = withLock s r $ rawGet s
  rawGets (AL s r) n = withLock s r $ rawGets s n
  rawGetContents (AL s r) = rawGetContents s -- locking is meaningless

instance (Lockable s, RetrievableStream a s) => RetrievableStream a (AutoLock a ma ba dest pos s) where
  rawGetUntil (AL s r) end = withLock s r $ rawGetUntil s end

instance (Lockable s, OStream a s) => OStream a (AutoLock a ma ba dest pos s) where
  rawPut (AL s r) e = withLock s r $ rawPut s e
  rawPuts (AL s r) es = withLock s r $ rawPuts s es

instance (Lockable s, IBlock ba s) => IBlock ba (AutoLock a ma ba dest pos s) where
  rawGetBufB (AL s r) bmode ptr sz = withLock s r $ rawGetBufB s bmode ptr sz

instance (Lockable s, OBlock ba s) => OBlock ba (AutoLock a ma ba dest pos s) where
  rawPutBufB (AL s r) bmode ptr sz = withLock s r $ rawPutBufB s bmode ptr sz

instance (Lockable s, IMemory ma s) => IMemory ma (AutoLock a ma ba dest pos s) where
  rawReadMem (AL s r) = withLock s r $ rawReadMem s
  rawEndRead (AL s r) p = withLock s r $ rawEndRead s p
  rawFillMemB (AL s r) bmode = withLock s r $ rawFillMemB s bmode

instance (Lockable s, OMemory ma s) => OMemory ma (AutoLock a ma ba dest pos s) where
  rawWriteMem (AL s r) = withLock s r $ rawWriteMem s
  rawEndWrite (AL s r) p = withLock s r $ rawEndWrite s p
  rawFlushMemB (AL s r) bmode = withLock s r $ rawFlushMemB s bmode
