{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Simple block buffering. Three transformers are provided for three
   different kinds of ports. Namely, input-only, output-only and
   input-output.
-}
module System.IO.SSC.Buffer
  ( iSizedBuffer
  , oSizedBuffer
  , ioSizedBuffer
  , iorSizedBuffer
  , iBuffer
  , oBuffer
  , ioBuffer
  , iorBuffer
  , IBuffer
  , OBuffer
  , IOBuffer
  , IORBuffer
  ) where

import System.IO.SSC.Classes
import System.IO.SSC.Internal.Ptr
import Control.Monad
import Data.IORef
import Data.Maybe
import Foreign.ForeignPtr
import Foreign.Marshal.Array
import Foreign.Storable
import Foreign.Ptr
import System.IO hiding (BufferMode)
import System.IO.Error
import System.IO.SSC.Internal.RawBuf
import System.IO.SSC.Internal.Debug
import System.IO.SSC.BufferSize

newtype IBuffer d pos a b = IBB (RawBuf a b)
newtype OBuffer d pos a b = OBB (RawBuf a b)
data IOBuffer d pos a b = IOBB (RawBuf a b) (IORef BufferMode)
-- | 'IORBuffer' is another type of buffer which deals with
-- redable-writable ports. Unlike IOBuffer, this transformer does
-- not depend on the underlying port to be RelativeSeekable.
-- The resulting port signals error if 
-- 'System.IO.SSC.Classes.bufFillB' is called when there are elements
-- left in the buffer. Consequently, it is not possible to layer any
-- non-trivial converter on top of the resulting port.
data IORBuffer d bpos a b = RBB (RawBuf a b) (IORef BufferMode) (IORef bpos)

data IORPos pos = IORPos pos Int

data BufferMode = BMWrite | BMRead
  deriving (Eq)

-- | Same as 'iSizedBuffer', but use the default buffer size.
iBuffer :: (Storable a) => b -> IO (IBuffer d pos a b)
iBuffer block = iSizedBuffer defaultBufferSize block

-- | Same as 'oSizedBuffer', but use the default buffer size.
oBuffer :: (Storable a) => b -> IO (OBuffer d pos a b)
oBuffer block = oSizedBuffer defaultBufferSize block

-- | Same as 'ioSizedBuffer', but use the default buffer size.
ioBuffer :: (Storable a) => b -> IO (IOBuffer d pos a b)
ioBuffer block = ioSizedBuffer defaultBufferSize block

-- | Same as 'iorSizedBuffer', but use the default buffer size.
iorBuffer :: (Storable a, Seekable pos b) => b -> IO (IORBuffer d pos a b)
iorBuffer block = iorSizedBuffer defaultBufferSize block


-- | Create memory IO port on top of input-only block IO interface,
-- using a buffer of the specified size.
iSizedBuffer :: (Storable a) => Int -> b -> IO (IBuffer d pos a b)
iSizedBuffer size block = liftM IBB $ newRawBuf size block

-- | Create memory IO port on top of output-only block IO interface,
-- using a buffer of the specified size.
oSizedBuffer :: (Storable a) => Int -> b -> IO (OBuffer d pos a b)
oSizedBuffer size block = liftM OBB $ newRawBuf size block

-- | Create memory IO port on top of input-output block IO interface,
-- using a buffer of the specified size.
ioSizedBuffer :: (Storable a) => Int -> b -> IO (IOBuffer d pos a b)
ioSizedBuffer size block = do
  rb <- newRawBuf size block
  rmode <- newIORef BMRead
  return (IOBB rb rmode)

-- | Create memory IO port on top of input-output block IO interface,
-- using a buffer of the specified size.
iorSizedBuffer :: (Storable a, Seekable pos b) => Int -> b -> IO (IORBuffer d pos a b)
iorSizedBuffer size block = do
  rb <- newRawBuf size block
  rmode <- newIORef BMRead
  pos <- newIORef =<< rawGetPos block
  return (RBB rb rmode pos)

bufFlush :: (OBlock a b) => BlockingMode -> RawBuf a b -> IO (Int, Bool)
bufFlush bmode rb@RB{ rbBase = base, rbBegin = begin, rbEnd = end, rbExtraObj = b} = do
  pbegin <- readIORef begin
  pend <- readIORef end
  msg $ "+bufFlush base=" ++ show base ++ "; size=" ++ show (distancePtr pend pbegin)
  let first = distancePtr pbegin base
  let last = distancePtr pend base
  result@(n, _) <- if first == last
    then return (0, False)
    else rawPutBufB b bmode pbegin (last - first)
  let first' = last - (first + n)
  moveArray base (advancePtr base (first + n)) first'
  let pbegin' = advancePtr base first'
  writeIORef begin base
  writeIORef end pbegin'
  rbTouch rb
  msg $ "-bufFlush result=" ++ show result
  return result

bufFlushCompletely :: (OBlock a b) => RawBuf a b -> IO ()
bufFlushCompletely rb = do
  bufFlush Blocking rb
  first <- readIORef (rbBegin rb)
  when (first /= rbBase rb) $ ioError $ mkIOError fullErrorType "Buffer: flush: buffer stuck" Nothing Nothing
  rbTouch rb

bufFill :: (IBlock a b) => BlockingMode -> RawBuf a b -> IO (Maybe Int)
bufFill bmode rb@RB{ rbBase = base, rbBegin = begin, rbEnd = end, rbEndBuf = bufend, rbExtraObj = b, rbEOI = reoi } = do
  eoi <- readIORef reoi
  if eoi
    then do
      writeIORef reoi False
      return Nothing
    else do
      first <- readIORef begin
      last <- readIORef end
      msg $ "+bufFill base=" ++ show base ++ "; first=" ++ show first ++ "; last=" ++ show last
      let r = distancePtr last first
      moveArray base first r
      let size = distancePtr bufend base
      let t = size - r
      result@(len, _) <- rawGetBufB b bmode (advancePtr base r) t
      writeIORef begin base
      writeIORef end (advancePtr base (r + len))
      rbTouch rb
      ret <- case result of
        (0, True) -> return Nothing
        (_, True) -> do
          writeIORef reoi True
          return (Just len)
        (_, False) -> return (Just len)
      msg $ "-bufFill last=" ++ show (advancePtr base (r + len)) ++ "+ ret=" ++ show ret
      return ret

bufFillBy :: (IBlock a b) => RawBuf a b -> Int -> IO ()
bufFillBy rb@RB{ rbBase = base, rbBegin = begin, rbEnd = end, rbEOI = reoi, rbExtraObj = b } oldused = do
  (len, eoi) <- rawGetBufB b Blocking base oldused
  when (len < oldused) $ ioError $ userError "System.IO.SSC.Buffer.bufFillBy: failed to revisit"
  writeIORef begin base
  writeIORef end (advancePtr base len)
  writeIORef reoi eoi
  rbTouch rb

bufDiscard :: (RelativeSeekable pos b, Storable a, Num pos) => RawBuf a b -> IO ()
bufDiscard rb = do
  used <- rbUsed rb
  rbTouch rb
  rawRelativeSeek (rbExtra rb) (fromIntegral $ negate used)

readModeGetPos :: (Seekable pos b, Storable a, Num pos) => RawBuf a b -> IO pos
readModeGetPos rb = do
  used <- rbUsed rb
  rbTouch rb
  raw_pos <- rawGetPos (rbExtra rb)
  return (raw_pos - fromIntegral used)

readModeNonRelativeSeek :: (Storable a) => RawBuf a b -> (b -> pos -> IO ()) -> pos -> IO ()
readModeNonRelativeSeek rb action pos = do
  let b = rbExtra rb
  action b pos
  rbReset rb
  rbTouch rb

readModeRelativeSeek :: (Storable a, RelativeSeekable pos b, Num pos) => RawBuf a b -> pos -> IO ()
readModeRelativeSeek rb amount = do
  let b = rbExtra rb
  used <- rbUsed rb
  rawRelativeSeek b (amount - fromIntegral used)
  rbReset rb
  rbTouch rb

writeModeGetPos :: (Seekable pos b, Storable a, Num pos) => RawBuf a b -> IO pos
writeModeGetPos rb = do
  raw_pos <- rawGetPos (rbExtra rb)
  used <- rbUsed rb
  rbTouch rb
  return (raw_pos + fromIntegral used)

writeModeSeek :: (Seekable pos b, OBlock a b) => RawBuf a b -> (b -> pos -> IO ()) -> pos -> IO ()
writeModeSeek rb action pos = do
  bufFlushCompletely rb
  action (rbExtra rb) pos
  rbReset rb
  rbTouch rb

prepareForRead :: (IBlock a b, Seekable pos b, OBlock a b) => IOBuffer d pos a b -> IO ()
prepareForRead (IOBB rb rmode) = do
  whenWriteMode rmode () $ bufFlushCompletely rb
  writeIORef rmode BMRead

prepareForWrite :: (IBlock a b, RelativeSeekable pos b, OBlock a b, Num pos) => IOBuffer d pos a b -> IO ()
prepareForWrite (IOBB rb rmode) = do
  whenReadMode rmode () $ bufDiscard rb
  writeIORef rmode BMWrite

resetLF :: (IBlock  a b, Seekable pos b, OBlock a b) => IORBuffer d pos a b -> IO ()
resetLF (RBB rb rmode rlf) = writeIORef rlf =<< rawGetPos (rbExtra rb)

prepareForReadR :: (IBlock a b, Seekable pos b, OBlock a b) => IORBuffer d pos a b -> IO ()
prepareForReadR rbb@(RBB rb rmode rlf) = do
  whenWriteMode rmode () $ bufFlushCompletely rb
  writeIORef rmode BMRead
  resetLF rbb

prepareForWriteR :: (IBlock a b, Seekable pos b, OBlock a b) => IORBuffer d pos a b -> IO ()
prepareForWriteR rbb@(RBB rb rmode pos) = do
  whenReadMode rmode () $ rDiscard rbb
  writeIORef rmode BMWrite

rDiscard :: (Seekable bpos b) => IORBuffer d bpos a b -> IO ()
rDiscard (RBB rb _ pos) = rawSetPos (rbExtra rb) =<< readIORef pos

whenReadMode :: IORef BufferMode -> a -> IO a -> IO a
whenReadMode rmode def action = do
  bmode <- readIORef rmode
  case bmode of
    BMRead -> action
    BMWrite -> return def

whenWriteMode :: IORef BufferMode -> a -> IO a -> IO a
whenWriteMode rmode def action = do
  bmode <- readIORef rmode
  case bmode of
    BMWrite -> action
    BMRead -> return def

instance (Port b) => Port (IBuffer d pos a b) where
  rawClose (IBB rb) = rawClose $ rbExtra rb
  rawIsClosed (IBB rb) = rawIsClosed $ rbExtra rb
  rawFlush _ = return ()

instance (OBlock a b) => Port (OBuffer d pos a b) where
  rawClose (OBB rb) = do
    bufFlushCompletely rb
    rawClose $ rbExtra rb
  rawIsClosed (OBB rb) = rawIsClosed $ rbExtra rb
  rawFlush (OBB rb) = do
    bufFlush Blocking rb
    rawFlush $ rbExtra rb

instance (OBlock a b) => Port (IOBuffer d pos a b) where
  rawClose (IOBB rb rmode) = do
    whenWriteMode rmode () $ bufFlushCompletely rb
    rawClose $ rbExtra rb
  rawIsClosed (IOBB rb _) = rawIsClosed $ rbExtra rb
  rawFlush (IOBB rb rmode) = do
    whenWriteMode rmode (0, False) $ bufFlush Blocking rb
    rawFlush $ rbExtra rb

instance (OBlock a b) => Port (IORBuffer d pos a b) where
  rawClose (RBB rb rmode _) = do
    whenWriteMode rmode () $ bufFlushCompletely rb
    rawClose $ rbExtra rb
  rawIsClosed (RBB rb _ _) = rawIsClosed $ rbExtra rb
  rawFlush (RBB rb rmode _) = do
    whenWriteMode rmode (0, False) $ bufFlush Blocking rb
    rawFlush $ rbExtra rb


instance (Destination d b) => Destination d (IBuffer d pos a b) where
  rawDest (IBB rb) = rawDest $ rbExtra rb

instance (Destination d b) => Destination d (OBuffer d pos a b) where
  rawDest (OBB rb) = rawDest $ rbExtra rb

instance (Destination d b) => Destination d (IOBuffer d pos a b) where
  rawDest (IOBB rb rmode) = rawDest $ rbExtra rb

instance (Destination d b) => Destination d (IORBuffer d pos a b) where
  rawDest (RBB rb rmode _) = rawDest $ rbExtra rb

instance (Lockable b) => Lockable (IBuffer d pos a b) where
  sLock (IBB rb) = sLock (rbExtra rb)
  sTryLock (IBB rb) = sTryLock (rbExtra rb)
  sUnsafeUnlock (IBB rb) = sUnsafeUnlock (rbExtra rb)

instance (Lockable b) => Lockable (OBuffer d pos a b) where
  sLock (OBB rb) = sLock (rbExtra rb)
  sTryLock (OBB rb) = sTryLock (rbExtra rb)
  sUnsafeUnlock (OBB rb) = sUnsafeUnlock (rbExtra rb)

instance (Lockable b) => Lockable (IOBuffer d pos a b) where
  sLock (IOBB rb _) = sLock (rbExtra rb)
  sTryLock (IOBB rb _) = sTryLock (rbExtra rb)
  sUnsafeUnlock (IOBB rb _) = sUnsafeUnlock (rbExtra rb)

instance (Lockable b) => Lockable (IORBuffer d pos a b) where
  sLock (RBB rb _ _) = sLock (rbExtra rb)
  sTryLock (RBB rb _ _) = sTryLock (rbExtra rb)
  sUnsafeUnlock (RBB rb _ _) = sUnsafeUnlock (rbExtra rb)

instance (Seekable pos b, IBlock a b, Num pos) => Seekable pos (IBuffer d pos a b) where
  rawGetPos (IBB rb) = readModeGetPos rb
  rawSetPos (IBB rb) = readModeNonRelativeSeek rb rawSetPos

instance (Seekable pos b, OBlock a b, Num pos) => Seekable pos (OBuffer d pos a b) where
  rawGetPos (OBB rb) = writeModeGetPos rb
  rawSetPos (OBB rb) = writeModeSeek rb rawSetPos

instance (Seekable pos b, IBlock a b, OBlock a b, Num pos) => Seekable pos (IOBuffer d pos a b) where
  rawGetPos (IOBB rb rmode) = do
    bmode <- readIORef rmode
    case bmode of
      BMRead -> readModeGetPos rb
      BMWrite -> writeModeGetPos rb
  rawSetPos (IOBB rb rmode) pos = do
    bmode <- readIORef rmode
    case bmode of
      BMRead -> readModeNonRelativeSeek rb rawSetPos pos
      BMWrite -> writeModeSeek rb rawSetPos pos

instance (Seekable pos b, IBlock a b, OBlock a b) => Seekable (IORPos pos) (IORBuffer d pos a b) where
  rawGetPos rbb@(RBB rb rmode rpos) = do
    prepareForReadR rbb
    used <- rbUsed rb
    rbTouch rb
    pos <- readIORef rpos
    return (IORPos pos used)
  rawSetPos rbb@(RBB rb rmode rpos) (IORPos oldp oldused) = do
    rawSetPos (rbExtra rb) oldp
    bufFillBy rb oldused
    writeIORef rpos oldp

instance (RelativeSeekable pos b, IBlock a b, Num pos) => RelativeSeekable pos (IBuffer d pos a b) where
  rawRelativeSeek (IBB rb) = readModeRelativeSeek rb

instance (RelativeSeekable pos b, OBlock a b, Num pos) => RelativeSeekable pos (OBuffer d pos a b) where
  rawRelativeSeek (OBB rb) = writeModeSeek rb rawRelativeSeek

instance (RelativeSeekable pos b, IBlock a b, OBlock a b, Num pos) => RelativeSeekable pos (IOBuffer d pos a b) where
  rawRelativeSeek (IOBB rb rmode) pos = do
    bmode <- readIORef rmode
    case bmode of
      BMRead -> readModeRelativeSeek rb pos
      BMWrite -> writeModeSeek rb rawRelativeSeek pos

instance (SizedInput pos b, IBlock a b, Num pos) => SizedInput pos (IBuffer d pos a b) where
  rawInputEnd (IBB rb) = rawInputEnd (rbExtra rb)
  rawSeekFromEnd (IBB rb) = readModeNonRelativeSeek rb rawSeekFromEnd

instance (SizedInput pos b, OBlock a b, Num pos) => SizedInput pos (OBuffer d pos a b) where
  rawInputEnd (OBB rb) = rawInputEnd (rbExtra rb)
  rawSeekFromEnd (OBB rb) pos = writeModeSeek rb rawSeekFromEnd pos

instance (SizedInput pos b, IBlock a b, OBlock a b, Num pos) => SizedInput pos (IOBuffer d pos a b) where
  rawInputEnd (IOBB rb rmode) = rawInputEnd (rbExtra rb)
  rawSeekFromEnd (IOBB rb rmode) pos = do
    bmode <- readIORef rmode
    case bmode of
      BMRead -> readModeNonRelativeSeek rb rawSeekFromEnd pos
      BMWrite -> writeModeSeek rb rawSeekFromEnd pos

instance (SizedOutput pos b, OBlock a b, Num pos) => SizedOutput pos (OBuffer d pos a b) where
  rawOutputEnd (OBB rb) = rawOutputEnd (rbExtra rb)

instance (SizedOutput pos b, OBlock a b, IBlock a b, Num pos) => SizedOutput pos (IOBuffer d pos a b) where
  rawOutputEnd (IOBB rb pos) = rawOutputEnd (rbExtra rb)

instance (IBlock a b, RelativeSeekable pos b, Num pos) => Discardable (IBuffer d pos a b) where
  rawDiscard (IBB rb) = bufDiscard rb

instance (OBlock a b, Seekable pos b) => Discardable (OBuffer d pos a b) where
  rawDiscard (OBB rb) = return ()

instance (IBlock a b, OBlock a b, RelativeSeekable pos b, Num pos) => Discardable (IOBuffer d pos a b) where
  rawDiscard (IOBB rb rmode) = whenReadMode rmode () $ bufDiscard rb

instance (IBlock a b, OBlock a b, Seekable pos b) => Discardable (IORBuffer d pos a b) where
  rawDiscard rbb@(RBB rb rmode _) = whenReadMode rmode () $ rDiscard rbb

instance (IBlock a b) => IMemory a (IBuffer d pos a b) where
  rawReadMem (IBB rb) = rbReadMem rb
  rawEndRead (IBB rb) endp = do
    rbEndRead rb endp
    rbTouch rb
  rawFillMemB (IBB rb) mode = bufFill mode rb

instance (IBlock a b, OBlock a b, Seekable pos b) => IMemory a (IOBuffer d pos a b) where
  rawReadMem iobb@(IOBB rb _) = do
    prepareForRead iobb
    rbReadMem rb
  rawEndRead iobb@(IOBB rb rmode) endp = do
    whenWriteMode rmode () $ ioError $ userError "System.IO.SSC.Buffer: rawEndRead: writing operation before terminating reading"
    rbEndRead rb endp
    rbTouch rb
  rawFillMemB iobb@(IOBB rb _) mode = do
    prepareForRead iobb
    bufFill mode rb

instance (IBlock a b, OBlock a b, Seekable pos b) => IMemory a (IORBuffer d pos a b) where
  rawReadMem rbb@(RBB rb _ _) = do
    prepareForReadR rbb
    rbReadMem rb
  rawEndRead rbb@(RBB rb rmode _) endp = do
    whenWriteMode rmode () $ ioError $ userError "System.IO.SSC.Buffer: rawEndRead: writing operation before terminating reading"
    rbEndRead rb endp
    rbTouch rb
  rawFillMemB rbb@(RBB rb _ rpos) mode = do
    prepareForReadR rbb
    empty <- rbEmpty rb
    when (not empty) $ ioError $ userError "System.IO.SSC.Buffer: rawFillMemB: Buffer not empty"
    r <- bufFill mode rb 
    writeIORef rpos =<< rawGetPos (rbExtra rb)
    return r

instance (OBlock a b) => OMemory a (OBuffer d pos a b) where
  rawWriteMem (OBB rb) = rbWriteMem rb
  rawEndWrite (OBB rb) endp = do
    rbEndWrite rb endp
    rbTouch rb
  rawFlushMemB (OBB rb) bmode = bufFlush bmode rb
  
instance (OBlock a b, IBlock a b, RelativeSeekable pos b, Num pos) => OMemory a (IOBuffer d pos a b) where
  rawWriteMem iobb@(IOBB rb _) = do
    prepareForWrite iobb
    rbWriteMem rb
  rawEndWrite iobb@(IOBB rb rmode) endp = do
    whenReadMode rmode () $ ioError $ userError "System.IO.SSC.Buffer: rawEndWrite: reading operation before terminating writing"
    rbEndWrite rb endp
    rbTouch rb
  rawFlushMemB (IOBB rb rmode) bmode = whenWriteMode rmode (0, False) $ bufFlush bmode rb

instance (OBlock a b, IBlock a b, Seekable pos b) => OMemory a (IORBuffer d pos a b) where
  rawWriteMem rbb@(RBB rb _ _) = do
    prepareForWriteR rbb
    rbWriteMem rb
  rawEndWrite rbb@(RBB rb rmode _) endp = do
    whenReadMode rmode () $ ioError $ userError "System.IO.SSC.Buffer: rawEndWrite: reading operation before terminating writing"
    rbEndWrite rb endp
    rbTouch rb
  rawFlushMemB (RBB rb rmode _) bmode = whenWriteMode rmode (0, False) $ bufFlush bmode rb

getBufB :: (IBlock a b) => RawBuf a b -> BlockingMode -> Ptr a -> Int -> IO (Int, Bool)
getBufB rb@RB{ rbBegin = begin, rbExtraObj = b } bmode buf size = do
  first <- readIORef begin
  used <- rbUsed rb
  r <- if size <= used
    then do
      copyArray buf first size
      if size == used
        then rbResetPtr rb
        else writeIORef begin (advancePtr first size)
      return (size, False)
    else do
      copyArray buf first used
      eoi <- readIORef (rbEOI rb)
      rbReset rb
      if eoi
        then return (used, True)
        else do 
          let 
            bmode' = if bmode == NonBlockingPreferred && used /= 0
              then NonBlocking
              else bmode
          (n, end) <- rawGetBufB b bmode' (advancePtr buf used) (size - used)
          return (n + used, end)
  rbTouch rb
  return r

putBufB :: (OBlock a b) => RawBuf a b -> BlockingMode -> Ptr a -> Int -> IO (Int, Bool)
putBufB rb@RB{ rbExtraObj = b } bmode buf size = do
  used <- rbUsed rb
  r <- case bmode of
    NonBlocking -> do
      r <- bufFlush NonBlocking rb
      case r of
        (n, False)
          | n == used -> rawPutBufB b NonBlocking buf size
        (_, end) -> return (0, end)
    _ -> do
      (_, end) <- bufFlush Blocking rb
      if end
        then return (0, True)
        else rawPutBufB b bmode buf size
  rbTouch rb
  return r

instance (IBlock a b) => IBlock a (IBuffer d pos a b) where
  rawGetBufB (IBB rb) = getBufB rb

instance (IBlock a b, Seekable pos b, OBlock a b) => IBlock a (IOBuffer d pos a b) where
  rawGetBufB iobb@(IOBB rb _) bmode buf size = do
    prepareForRead iobb
    getBufB rb bmode buf size

instance (IBlock a b, Seekable pos b, OBlock a b) => IBlock a (IORBuffer d pos a b) where
  rawGetBufB rbb@(RBB rb _ _) bmode buf size = do
    prepareForReadR rbb
    getBufB rb bmode buf size


instance (OBlock a b) => OBlock a (OBuffer d pos a b) where
  rawPutBufB (OBB rb) = putBufB rb

instance (OBlock a b, RelativeSeekable pos b, Num pos, IBlock a b) => OBlock a (IOBuffer d pos a b) where
  rawPutBufB iobb@(IOBB rb _) bmode buf size = do
    prepareForWrite iobb
    putBufB rb bmode buf size

instance (OBlock a b, Seekable pos b, IBlock a b) => OBlock a (IORBuffer d pos a b) where
  rawPutBufB rbb@(RBB rb _ _) bmode buf size = do
    prepareForWriteR rbb
    putBufB rb bmode buf size
