{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   This module defines a number of classes, each of which describes a
   capability (readable, writable, supports block IO, ...) of
   Handle-like types. All of these types are supposed to be an
   instance of 'Port' class.
-}
module System.IO.SSC.Classes
  ( BlockingMode(..)
  , Port(..)
  , Destination(..)
  , Lockable(..)
  , Discardable(..)
  , Seekable(..)
  , RelativeSeekable(..)
  , SizedInput(..)
  , SizedOutput(..)
  , IStream(..)
  , RetrievableStream(..)
  , OStream(..)
  , IBlock(..)
  , OBlock(..)
  , IMemory(..)
  , OMemory(..)
  ) where

import Control.Concurrent.MVar
import Data.Word
import System.IO
import Foreign.Storable
import Foreign.Ptr

-- | `BlockingMode` is used to specify when an operation can block.
data BlockingMode
  = NonBlocking -- ^ Operation can never block.
  | Blocking 
    -- ^ Operation may block. Operation should not return before the
    -- specified number of elements are read\/written, the end of
    -- input is reached, or an error is ocurred.
  | NonBlockingPreferred
    -- ^ If some data can be read\/written without blocking, behave
    -- as `NonBlocking` is specified. Otherwise, behave as
    -- `Blocking`.
  deriving (Show, Eq)

-- | The 'Port' class represents something through which data can be
-- passed to\/from somewhere.
class Port s where
  -- | Close the port. This may not perform necessary cleanups, such
  -- as flushing buffers. Use 'System.IO.SSC.Utils.rawClose' for usual
  -- close operation.
  --
  -- After this operation, any operation on the port will either be
  -- silently ignored, or raises an error.
  rawClose :: s -> IO ()
  -- | Test whether the port is closed.
  rawIsClosed :: s -> IO Bool
  -- | Flush all the output buffers.
  rawFlush :: s -> IO ()

-- | Ports that is connected to somewhere.
class Destination dest s | s -> dest where
  -- | Retrieve information about the destination of the port. The
  -- return type depends on the type of the port being used.
  rawDest :: s -> IO dest

-- | Represents a port that supports explicit locking\/unlocking. 
-- Basically all ports should be lockable. Being an instance of
-- 'Lockable' does not mean its operations are automatically
-- guarded with a lock.
class Lockable s where
  sLock :: s -> IO ()
  sTryLock :: s -> IO Bool
  sUnsafeUnlock :: s -> IO ()

-- | Ports that can discard its input buffers.
class (Port s) => Discardable s where
  -- | Discard all the input buffers.
  rawDiscard :: s -> IO ()

-- | Ports that support position restore.
class (Port s) => Seekable pos s | s -> pos where
  rawGetPos :: s -> IO pos
  rawSetPos :: s -> pos -> IO ()

-- | Ports that support relative seek.
class (Seekable pos s) => RelativeSeekable pos s where
  rawRelativeSeek :: s -> pos -> IO ()

-- | Ports that support seek operation relative to the end of input.
class (Seekable pos s) => SizedInput pos s where
  rawInputEnd :: s -> IO pos
  rawSeekFromEnd :: s -> pos -> IO ()

-- | Ports which have constant size limit on output.
class (Seekable pos s) => SizedOutput pos s where
  rawOutputEnd :: s -> IO pos

-- | Element-wise input.
class (Port s) => IStream a s | s -> a where
  -- | Get one element from the stream.
  rawGet :: s -> IO a
  -- | Get specified number of elements. If the end of input is
  -- reached, elements read so far are returned. If no elements
  -- are read, end-of-file exception is raised.
  rawGets :: s -> Int -> IO [a]
  -- | Get the whole input as one list, reading lazily. After
  -- this operation, any operation on the same port, including
  -- 'rawClose', will be invalid.
  rawGetContents :: s -> IO [a]

-- | Streams that supports 'rawGetUntil'. This method belongs to the
-- separate class because it may require the element type to have
-- additional property (such as equality).
class (IStream a s) => RetrievableStream a s | s -> a where
  -- | Get elements from the stream until the specified element is
  -- encountered.
  rawGetUntil :: s -> a -> IO [a]

-- | Element-wise output.
class (Port s) => OStream a s | s -> a where
  -- | Put one element to the stream.
  rawPut :: s -> a -> IO ()
  -- | Put several elements to the stream.
  rawPuts :: s -> [a] -> IO ()

-- | Buffer-based input where the user provides the buffer.
class (Port s, Storable a) => IBlock a s | s -> a where
  -- | Read elements and store it in the specified buffer, with
  -- specified blocking policy. It returns number of elements
  -- actually read, and whether the end of input is reached.
  rawGetBufB :: s -> BlockingMode -> Ptr a -> Int -> IO (Int, Bool)

-- | Buffer-based output where the user provides the buffer.
class (Port s, Storable a) => OBlock a s | s -> a where
  -- | Write elements stored in the buffer, with specified blocking
  -- policy. It returns number of elements actually written, and
  -- whether the end of output is reached.
  rawPutBufB :: s -> BlockingMode -> Ptr a -> Int -> IO (Int, Bool)

-- | Buffer-based input where the port provides the buffer.
class (Port b, Storable a) => IMemory a b | b -> a where
  rawReadMem :: b -> IO (Ptr a, Ptr a)
  rawEndRead :: b -> Ptr a -> IO ()
  -- | Extend the internal memory by filling it with new elements.
  -- This operation returns number of new elements, or Nothing if the
  -- end of input is reached before any elements are read.
  rawFillMemB :: b -> BlockingMode -> IO (Maybe Int)

-- | Buffer-based output where the port provides the buffer.
class (Port b, Storable a) => OMemory a b | b -> a where
  rawWriteMem :: b -> IO (Ptr a, Ptr a)
  rawEndWrite :: b -> Ptr a -> IO ()
  -- | Extend the internal memory so that new elements can be
  -- written to it. This operation returns number of elements
  -- written, and whether the end of output is eached.
  rawFlushMemB :: b -> BlockingMode -> IO (Int, Bool)
 
