{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental
   
   UTF-8 conversion.
-}
module System.IO.SSC.UTF8Block
  ( utf8ToCharConverter
  , charToUTF8Converter
  , utf8ToUTF16Converter
  , utf16ToUTF8Converter
  , iUTF8Block
  , oUTF8Block
  , ioUTF8Block
  ) where

import Control.Monad
import Data.Bits
import Data.Word
import Data.Char
import Foreign.Storable
import Foreign.Marshal.Array
import Foreign.Ptr

import System.IO.SSC.Convert

-- See rfc2279

utf8ToCharConverter :: Converter Word8 Char
utf8ToCharConverter iptr isize optr osize = loop 0 0
  where
    loop m n
      | m == isize || n == osize = return (m, n)
      | otherwise = do
      r <- readUTF8 (advancePtr iptr m) (isize - m)
      case r of
        Success sz c -> do
          pokeElemOff optr n (chr c)
          loop (m + sz) (n + 1)
        Failure -> return (m, n)

utf8ToUTF16Converter :: Converter Word8 Word16
utf8ToUTF16Converter iptr isize optr osize = loop 0 0
  where
    loop m n
      | m == isize || n == osize = return (m, n)
      | otherwise = do
        r <- readUTF8 (advancePtr iptr m) (isize - m)
        case r of
          Success sz c
            | 0x10000 <= c -> if osize <= 1 then return (m, n) else do
              pokeElemOff optr n (fromIntegral $ ((c `shiftR` 10) .&. 0x3ff) .|. 0xd800)
              pokeElemOff optr (n + 1) (fromIntegral $ (c .&. 0x3ff) .|. 0xdc00)
              loop (m + sz) (n + 2)
            | otherwise -> do
              when (0xd800 <= c && c < 0xe000) ilseq
              pokeElemOff optr n (fromIntegral c)
              loop (m + sz) (n + 1)
          Failure -> return (m, n)

charToUTF8Converter :: Converter Char Word8
charToUTF8Converter iptr isize optr osize = loop 0 0
  where
    loop m n
      | m == isize || n == osize = return (m, n)
      | otherwise = do
        c <- peekElemOff iptr m
        sz <- writeUTF8 (advancePtr optr n) (osize - n) (ord c)
        case sz of
          0 -> return (m, n)
          _ -> loop (m + 1) (n + sz)

utf16ToUTF8Converter :: Converter Word16 Word8
utf16ToUTF8Converter iptr isize optr osize = loop 0 0
  where
    loop m n
      | m == isize || n == osize = return (m, n)
      | otherwise = do
        c0 <- peekElemOff iptr m
        if c0 < 0xd800 || 0xe000 <= c0
          then write (int c0) 1
          else do
            when (0xdc00 <= c0) ilseq
            if m + 1 >= isize
              then return (m, n)
              else do
                c1 <- peekElemOff iptr (m + 1)
                when (c1 < 0xdc00 || 0xe000 <= c1) ilseq
                write (((int c0 .&. 0x3ff) `shiftL` 10) .|. (int c1 .&. 0x3ff)) 2
      where
        write c ni = do
          sz <- writeUTF8 (advancePtr optr n) (osize - n) c
          case sz of
            0 -> return (m, n)
            _ -> loop (m + ni) (n + sz)
                

data Result 
  = Success !Int !Int
  | Failure

readUTF8 :: Ptr Word8 -> Int -> IO Result
readUTF8 p size = do
  b0 <- peek p
  case () of
    _ | b0 < 0x80 -> do
        return $ Success 1 (fromIntegral b0)
      | b0 < 0xc0 -> ilseq
      | b0 < 0xe0 -> if size < 2 then return Failure else do
        b1 <- peekElemOff p 1
        check b1
        let c = ((int b0 .&. 0x1f) `shiftL` 6) .|. int b1 .&. 0x3f
        when (c < 0x80) ilseq
        return $ Success 2 c
      | b0 < 0xf0 -> if size < 3 then return Failure else do
        b1 <- peekElemOff p 1
        b2 <- peekElemOff p 2
        check b1
        check b2
        let 
          c = ((int b0 .&. 0xf) `shiftL` 12)
            .|. ((int b1 .&. 0x3f) `shiftL` 6)
            .|. (int b2 .&. 0x3f)
        when (c < 0x0800) ilseq
        return $ Success 3 c
      | b0 < 0xf8 -> if size < 4 then return Failure else do
        b1 <- peekElemOff p 1
        b2 <- peekElemOff p 2
        b3 <- peekElemOff p 3
        check b1
        check b2
        check b3
        let
          c = ((int b0 .&. 0x7) `shiftL` 18)
            .|. ((int b1 .&. 0x3f) `shiftL` 12)
            .|. ((int b2 .&. 0x3f) `shiftL` 6)
            .|. (int b3 .&. 0x3f)
        when (c < 0x10000) ilseq
        return $ Success 4 c
      | otherwise -> ilseq
  
writeUTF8 :: Ptr Word8 -> Int -> Int -> IO Int
writeUTF8 ptr size c
  | c < 0x80 = do
    p 0 (int c)
    return 1
  | c < 0x800 = if size < 2 then return 0 else do
    p 0 (0xc0 .|. (c `shiftR` 6))
    p 1 (0x80 .|. (c .&. 0x3f))
    return 2
  | c < 0x10000 = if size < 3 then return 0 else do
    p 0 (0xe0 .|. (c `shiftR` 12))
    p 1 (0x80 .|. ((c `shiftR` 6) .&. 0x3f))
    p 2 (0x80 .|. (c .&. 0x3f))
    return 3
  | otherwise = if size < 4 then return 0 else do
    p 0 (0xf0 .|. (c `shiftR` 18))
    p 1 (0x80 .|. ((c `shiftR` 12) .&. 0x3f))
    p 2 (0x80 .|. ((c `shiftR` 6) .&. 0x3f))
    p 3 (0x80 .|. (c .&. 0x3f))
    return 4
  where
    p off w = pokeElemOff ptr off (int w)

check :: Word8 -> IO ()
check w = when (w < 0x80 || 0xc0 <= w) ilseq 

ilseq = ioError $ userError $ "System.IO.SSC.UTF8Block: Illegal UTF-8 sequence"

int :: (Integral a, Num b) => a -> b
int = fromIntegral

iUTF8Block :: m -> IO (IConversionBlock d pos m a Word8 Char)
iUTF8Block mem = iConversionBlock utf8ToCharConverter mem

oUTF8Block :: m -> IO (OConversionBlock d pos m a Word8 Char)
oUTF8Block mem = oConversionBlock charToUTF8Converter mem

ioUTF8Block :: m -> IO (IOConversionBlock d pos m a Word8 Char)
ioUTF8Block mem = ioConversionBlock utf8ToCharConverter charToUTF8Converter mem
